# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains toy objective functions, useful for experimentation. A number of them have been
taken from `this Virtual Library of Simulation Experiments
<https://web.archive.org/web/20211015101644/https://www.sfu.ca/~ssurjano/> (:cite:`ssurjano2021`)`_.
"""

from __future__ import annotations

import math
from dataclasses import dataclass
from math import pi
from typing import Callable, Generic, Sequence

import tensorflow as tf
from check_shapes import check_shapes

from ..space import Box, Constraint, LinearConstraint, NonlinearConstraint, SearchSpaceType
from ..types import TensorType

ObjectiveTestFunction = Callable[[TensorType], TensorType]
"""A synthetic test function"""


@dataclass(frozen=True)
class ObjectiveTestProblem(Generic[SearchSpaceType]):
    """
    Convenience container class for synthetic objective test functions.
    """

    name: str
    """The test function name"""

    objective: ObjectiveTestFunction
    """The synthetic test function"""

    search_space: SearchSpaceType
    """The (continuous) search space of the test function"""

    @property
    def dim(self) -> int:
        """The input dimensionality of the test function"""
        return self.search_space.dimension

    @property
    def bounds(self) -> list[list[float]]:
        """The input space bounds of the test function"""
        return [self.search_space.lower, self.search_space.upper]


@dataclass(frozen=True)
class SingleObjectiveTestProblem(ObjectiveTestProblem[SearchSpaceType]):
    """
    Convenience container class for synthetic single-objective test functions,
    including the global minimizers and minimum.
    """

    minimizers: TensorType
    """The global minimizers of the test function."""

    minimum: TensorType
    """The global minimum of the test function."""


def check_objective_shapes(d: int) -> Callable[[ObjectiveTestFunction], ObjectiveTestFunction]:
    """Returns a decorator for checking the shape of single objective test functions."""
    return check_shapes(f"x: [batch..., {d}]", "return: [batch..., 1]")


def _branin_internals(x: TensorType, scale: TensorType, translate: TensorType) -> TensorType:
    x0 = x[..., :1] * 15.0 - 5.0
    x1 = x[..., 1:] * 15.0

    b = 5.1 / (4 * math.pi**2)
    c = 5 / math.pi
    r = 6
    s = 10
    t = 1 / (8 * math.pi)

    return scale * ((x1 - b * x0**2 + c * x0 - r) ** 2 + s * (1 - t) * tf.cos(x0) + translate)


@check_objective_shapes(d=2)
def branin(x: TensorType) -> TensorType:
    """
    The Branin-Hoo function over :math:`[0, 1]^2`. See
    :cite:`Picheny2013` for details.

    :param x: The points at which to evaluate the function, with shape [..., 2].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    return _branin_internals(x, 1, 10)


@check_objective_shapes(d=2)
def scaled_branin(x: TensorType) -> TensorType:
    """
    The Branin-Hoo function, rescaled to have zero mean and unit variance over :math:`[0, 1]^2`. See
    :cite:`Picheny2013` for details.

    :param x: The points at which to evaluate the function, with shape [..., 2].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    return _branin_internals(x, 1 / 51.95, -44.81)


_ORIGINAL_BRANIN_MINIMIZERS = tf.constant(
    [[-math.pi, 12.275], [math.pi, 2.275], [9.42478, 2.475]], tf.float64
)

Branin = SingleObjectiveTestProblem(
    name="Branin",
    objective=branin,
    search_space=Box([0.0], [1.0]) ** 2,
    minimizers=(_ORIGINAL_BRANIN_MINIMIZERS + [5.0, 0.0]) / 15.0,
    minimum=tf.constant([0.397887], tf.float64),
)
"""The Branin-Hoo function over :math:`[0, 1]^2`. See :cite:`Picheny2013` for details."""

ScaledBranin = SingleObjectiveTestProblem(
    name="Scaled Branin",
    objective=scaled_branin,
    search_space=Branin.search_space,
    minimizers=Branin.minimizers,
    minimum=tf.constant([-1.047393], tf.float64),
)
"""The Branin-Hoo function, rescaled to have zero mean and unit variance over :math:`[0, 1]^2`. See
:cite:`Picheny2013` for details."""


def _scaled_branin_constraints() -> Sequence[Constraint]:
    def _nlc_func0(x: TensorType) -> TensorType:
        c0 = x[..., 0] - 0.2 - tf.sin(x[..., 1])
        c0 = tf.expand_dims(c0, axis=-1)
        return c0

    def _nlc_func1(x: TensorType) -> TensorType:
        c1 = x[..., 0] - tf.cos(x[..., 1])
        c1 = tf.expand_dims(c1, axis=-1)
        return c1

    constraints: Sequence[Constraint] = [
        LinearConstraint(
            A=tf.constant([[-1.0, 1.0], [1.0, 0.0], [0.0, 1.0]]),
            lb=tf.constant([-0.4, 0.15, 0.2]),
            ub=tf.constant([0.5, 0.9, 0.9]),
        ),
        NonlinearConstraint(_nlc_func0, tf.constant(-1.0), tf.constant(0.0)),
        NonlinearConstraint(_nlc_func1, tf.constant(-0.8), tf.constant(0.0)),
    ]

    return constraints


ConstrainedScaledBranin = SingleObjectiveTestProblem(
    name="Constrained Scaled Branin",
    objective=scaled_branin,
    search_space=Box(
        Branin.search_space.lower,
        Branin.search_space.upper,
        constraints=_scaled_branin_constraints(),
    ),
    minimizers=tf.constant([[0.16518, 0.66518]], tf.float64),
    minimum=tf.constant([-0.99888], tf.float64),
)
"""The rescaled Branin-Hoo function with a combination of linear and nonlinear constraints on the
search space."""


@check_objective_shapes(d=2)
def simple_quadratic(x: TensorType) -> TensorType:
    """
    A trivial quadratic function over :math:`[0, 1]^2`. Useful for quick testing.

    :param x: The points at which to evaluate the function, with shape [..., 2].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    return -tf.math.reduce_sum(x, axis=-1, keepdims=True) ** 2


SimpleQuadratic = SingleObjectiveTestProblem(
    name="Simple Quadratic",
    objective=simple_quadratic,
    search_space=Branin.search_space,
    minimizers=tf.constant([[1.0, 1.0]], tf.float64),
    minimum=tf.constant([-4.0], tf.float64),
)
"""A trivial quadratic function over :math:`[0, 1]^2`. Useful for quick testing."""


@check_objective_shapes(d=1)
def gramacy_lee(x: TensorType) -> TensorType:
    """
    The Gramacy & Lee function, typically evaluated over :math:`[0.5, 2.5]`. See
    :cite:`gramacy2012cases` for details.

    :param x: Where to evaluate the function, with shape [..., 1].
    :return: The function values, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    return tf.sin(10 * math.pi * x) / (2 * x) + (x - 1) ** 4


GramacyLee = SingleObjectiveTestProblem(
    name="Gramacy & Lee",
    objective=gramacy_lee,
    search_space=Box([0.5], [2.5]),
    minimizers=tf.constant([[0.548562]], tf.float64),
    minimum=tf.constant([-0.869011], tf.float64),
)
"""The Gramacy & Lee function, typically evaluated over :math:`[0.5, 2.5]`. See
:cite:`gramacy2012cases` for details."""


@check_objective_shapes(d=2)
def logarithmic_goldstein_price(x: TensorType) -> TensorType:
    """
    A logarithmic form of the Goldstein-Price function, with zero mean and unit variance over
    :math:`[0, 1]^2`. See :cite:`Picheny2013` for details.

    :param x: The points at which to evaluate the function, with shape [..., 2].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    x0, x1 = tf.split(4 * x - 2, 2, axis=-1)

    a = (x0 + x1 + 1) ** 2
    b = 19 - 14 * x0 + 3 * x0**2 - 14 * x1 + 6 * x0 * x1 + 3 * x1**2
    c = (2 * x0 - 3 * x1) ** 2
    d = 18 - 32 * x0 + 12 * x0**2 + 48 * x1 - 36 * x0 * x1 + 27 * x1**2

    return (1 / 2.427) * (tf.math.log((1 + a * b) * (30 + c * d)) - 8.693)


LogarithmicGoldsteinPrice = SingleObjectiveTestProblem(
    name="Logarithmic Goldstein-Price",
    objective=logarithmic_goldstein_price,
    search_space=Box([0.0], [1.0]) ** 2,
    minimizers=tf.constant([[0.5, 0.25]], tf.float64),
    minimum=tf.constant([-3.12913], tf.float64),
)
"""A logarithmic form of the Goldstein-Price function, with zero mean and unit variance over
:math:`[0, 1]^2`. See :cite:`Picheny2013` for details."""


@check_objective_shapes(d=3)
def hartmann_3(x: TensorType) -> TensorType:
    """
    The Hartmann 3 test function over :math:`[0, 1]^3`. This function has 3 local
    and one global minima. See https://www.sfu.ca/~ssurjano/hart3.html for details.

    :param x: The points at which to evaluate the function, with shape [..., 3].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    a = [1.0, 1.2, 3.0, 3.2]
    A = [[3.0, 10.0, 30.0], [0.1, 10.0, 35.0], [3.0, 10.0, 30.0], [0.1, 10.0, 35.0]]
    P = [
        [0.3689, 0.1170, 0.2673],
        [0.4699, 0.4387, 0.7470],
        [0.1091, 0.8732, 0.5547],
        [0.0381, 0.5743, 0.8828],
    ]

    inner_sum = -tf.reduce_sum(A * (tf.expand_dims(x, 1) - P) ** 2, -1)
    return -tf.reduce_sum(a * tf.math.exp(inner_sum), -1, keepdims=True)


Hartmann3 = SingleObjectiveTestProblem(
    name="Hartmann 3",
    objective=hartmann_3,
    search_space=Box([0.0], [1.0]) ** 3,
    minimizers=tf.constant([[0.114614, 0.555649, 0.852547]], tf.float64),
    minimum=tf.constant([-3.86278], tf.float64),
)
"""The Hartmann 3 test function over :math:`[0, 1]^3`. This function has 3 local
and one global minima. See https://www.sfu.ca/~ssurjano/hart3.html for details."""


@check_objective_shapes(d=4)
def shekel_4(x: TensorType) -> TensorType:
    """
    The Shekel test function over :math:`[0, 1]^4`. This function has ten local
    minima and a single global minimum. See https://www.sfu.ca/~ssurjano/shekel.html for details.
    Note that we rescale the original problem, which is typically defined
    over `[0, 10]^4`.

    :param x: The points at which to evaluate the function, with shape [..., 4].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    y: TensorType = x * 10.0

    beta = [0.1, 0.2, 0.2, 0.4, 0.4, 0.6, 0.3, 0.7, 0.5, 0.5]
    C = [
        [4.0, 1.0, 8.0, 6.0, 3.0, 2.0, 5.0, 8.0, 6.0, 7.0],
        [4.0, 1.0, 8.0, 6.0, 7.0, 9.0, 3.0, 1.0, 2.0, 3.6],
        [4.0, 1.0, 8.0, 6.0, 3.0, 2.0, 5.0, 8.0, 6.0, 7.0],
        [4.0, 1.0, 8.0, 6.0, 7.0, 9.0, 3.0, 1.0, 2.0, 3.6],
    ]

    inner_sum = tf.reduce_sum((tf.expand_dims(y, -1) - C) ** 2, 1)
    inner_sum += tf.cast(tf.transpose(beta), dtype=inner_sum.dtype)
    return -tf.reduce_sum(inner_sum ** (-1), -1, keepdims=True)


Shekel4 = SingleObjectiveTestProblem(
    name="Shekel 4",
    objective=shekel_4,
    search_space=Box([0.0], [1.0]) ** 4,
    minimizers=tf.constant([[0.4, 0.4, 0.4, 0.4]], tf.float64),
    minimum=tf.constant([-10.5363], tf.float64),
)
"""The Shekel test function over :math:`[0, 1]^4`. This function has ten local
minima and a single global minimum. See https://www.sfu.ca/~ssurjano/shekel.html for details.
Note that we rescale the original problem, which is typically defined
over `[0, 10]^4`."""


def levy(x: TensorType, d: int) -> TensorType:
    """
    The Levy test function over :math:`[0, 1]^d`. This function has many local
    minima and a single global minimum. See https://www.sfu.ca/~ssurjano/levy.html for details.
    Note that we rescale the original problem, which is typically defined
    over `[-10, 10]^d`, to be defined over a unit hypercube :math:`[0, 1]^d`.

    :param x: The points at which to evaluate the function, with shape [..., d].
    :param d: The dimension of the function.
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    tf.debugging.assert_greater_equal(d, 1)
    tf.debugging.assert_shapes([(x, (..., d))])

    w: TensorType = 1 + ((x * 20.0 - 10) - 1) / 4
    term1 = tf.pow(tf.sin(pi * w[..., 0:1]), 2)
    term3 = (w[..., -1:] - 1) ** 2 * (1 + tf.pow(tf.sin(2 * pi * w[..., -1:]), 2))
    wi = w[..., 0:-1]
    wi_sum = tf.reduce_sum(
        (wi - 1) ** 2 * (1 + 10 * tf.pow(tf.sin(pi * wi + 1), 2)), axis=-1, keepdims=True
    )
    return term1 + wi_sum + term3


@check_objective_shapes(d=8)
def levy_8(x: TensorType) -> TensorType:
    """
    Convenience function for the 8-dimensional :func:`levy` function, with output
    normalised to unit interval

    :param x: The points at which to evaluate the function, with shape [..., 8].
    :return: The function values at ``x``, with shape [..., 1].
    """
    return levy(x, d=8) / 450.0


Levy8 = SingleObjectiveTestProblem(
    name="Levy 8",
    objective=levy_8,
    search_space=Box([0.0], [1.0]) ** 8,
    minimizers=tf.constant([[11 / 20] * 8], tf.float64),
    minimum=tf.constant([0], tf.float64),
)
"""Convenience function for the 8-dimensional :func:`levy` function.
Taken from https://www.sfu.ca/~ssurjano/levy.html"""


def rosenbrock(x: TensorType, d: int) -> TensorType:
    """
    The Rosenbrock function, also known as the Banana function, is a unimodal function,
    however the minima lies in a narrow valley. Even though this valley is
    easy to find, convergence to the minimum is difficult. See
    https://www.sfu.ca/~ssurjano/rosen.html for details. Inputs are rescaled to
    be defined over a unit hypercube :math:`[0, 1]^d`.

    :param x: The points at which to evaluate the function, with shape [..., d].
    :param d: The dimension of the function.
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    tf.debugging.assert_greater_equal(d, 1)
    tf.debugging.assert_shapes([(x, (..., d))])

    y: TensorType = x * 15.0 - 5
    unscaled_function = tf.reduce_sum(
        (100.0 * (y[..., 1:] - y[..., :-1]) ** 2 + (1 - y[..., :-1]) ** 2), axis=-1, keepdims=True
    )
    return unscaled_function


@check_objective_shapes(d=4)
def rosenbrock_4(x: TensorType) -> TensorType:
    """
    Convenience function for the 4-dimensional :func:`rosenbrock` function with steepness 10.
    It is rescaled to have zero mean and unit variance over :math:`[0, 1]^4. See
    :cite:`Picheny2013` for details.

    :param x: The points at which to evaluate the function, with shape [..., 4].
    :return: The function values at ``x``, with shape [..., 1].
    """
    return (rosenbrock(x, d=4) - 3.827 * 1e5) / (3.755 * 1e5)


Rosenbrock4 = SingleObjectiveTestProblem(
    name="Rosenbrock 4",
    objective=rosenbrock_4,
    search_space=Box([0.0], [1.0]) ** 4,
    minimizers=tf.constant([[0.4] * 4], tf.float64),
    minimum=tf.constant([-1.01917], tf.float64),
)
"""The Rosenbrock function, rescaled to have zero mean and unit variance over :math:`[0, 1]^4. See
:cite:`Picheny2013` for details.
This function (also known as the Banana function) is unimodal, however the minima
lies in a narrow valley."""


@check_objective_shapes(d=5)
def ackley_5(x: TensorType) -> TensorType:
    """
    The Ackley test function over :math:`[0, 1]^5`. This function has
    many local minima and a global minima. See https://www.sfu.ca/~ssurjano/ackley.html
    for details.
    Note that we rescale the original problem, which is typically defined
    over `[-32.768, 32.768]`.

    :param x: The points at which to evaluate the function, with shape [..., 5].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    x = (x - 0.5) * (32.768 * 2.0)

    exponent_1 = -0.2 * tf.math.sqrt((1 / 5.0) * tf.reduce_sum(x**2, -1))
    exponent_2 = (1 / 5.0) * tf.reduce_sum(tf.math.cos(2.0 * math.pi * x), -1)

    function = (
        -20.0 * tf.math.exp(exponent_1)
        - tf.math.exp(exponent_2)
        + 20.0
        + tf.cast(tf.math.exp(1.0), dtype=x.dtype)
    )

    return tf.expand_dims(function, -1)


Ackley5 = SingleObjectiveTestProblem(
    name="Ackley 5",
    objective=ackley_5,
    search_space=Box([0.0], [1.0]) ** 5,
    minimizers=tf.constant([[0.5, 0.5, 0.5, 0.5, 0.5]], tf.float64),
    minimum=tf.constant([0.0], tf.float64),
)
"""The Ackley test function over :math:`[0, 1]^5`. This function has
many local minima and a global minima. See https://www.sfu.ca/~ssurjano/ackley.html
for details.
Note that we rescale the original problem, which is typically defined
over `[-32.768, 32.768]`."""


@check_objective_shapes(d=6)
def hartmann_6(x: TensorType) -> TensorType:
    """
    The Hartmann 6 test function over :math:`[0, 1]^6`. This function has
    6 local and one global minima. See https://www.sfu.ca/~ssurjano/hart6.html
    for details.

    :param x: The points at which to evaluate the function, with shape [..., 6].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    a = [1.0, 1.2, 3.0, 3.2]
    A = [
        [10.0, 3.0, 17.0, 3.5, 1.7, 8.0],
        [0.05, 10.0, 17.0, 0.1, 8.0, 14.0],
        [3.0, 3.5, 1.7, 10.0, 17.0, 8.0],
        [17.0, 8.0, 0.05, 10.0, 0.1, 14.0],
    ]
    P = [
        [0.1312, 0.1696, 0.5569, 0.0124, 0.8283, 0.5886],
        [0.2329, 0.4135, 0.8307, 0.3736, 0.1004, 0.9991],
        [0.2348, 0.1451, 0.3522, 0.2883, 0.3047, 0.6650],
        [0.4047, 0.8828, 0.8732, 0.5743, 0.1091, 0.0381],
    ]

    inner_sum = -tf.reduce_sum(A * (tf.expand_dims(x, 1) - P) ** 2, -1)
    return -tf.reduce_sum(a * tf.math.exp(inner_sum), -1, keepdims=True)


Hartmann6 = SingleObjectiveTestProblem(
    name="Hartmann 6",
    objective=hartmann_6,
    search_space=Box([0.0], [1.0]) ** 6,
    minimizers=tf.constant([[0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573]], tf.float64),
    minimum=tf.constant([-3.32237], tf.float64),
)
"""The Hartmann 6 test function over :math:`[0, 1]^6`. This function has
6 local and one global minima. See https://www.sfu.ca/~ssurjano/hart6.html
for details."""


def michalewicz(x: TensorType, d: int = 2, m: int = 10) -> TensorType:
    """
    The Michalewicz function over :math:`[0, \\pi]` for all i=1,...,d. Dimensionality is determined
    by the parameter ``d`` and it features steep ridges and drops. It has :math:`d!` local minima,
    and it is multimodal. The parameter ``m`` defines the steepness of they valleys and ridges; a
    larger ``m`` leads to a more difficult search. The recommended value of ``m`` is 10. See
    https://www.sfu.ca/~ssurjano/egg.html for details.

    :param x: The points at which to evaluate the function, with shape [..., d].
    :param d: The dimension of the function.
    :param m: The steepness of the valleys/ridges.
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    tf.debugging.assert_greater_equal(d, 1)
    tf.debugging.assert_shapes([(x, (..., d))])

    xi = tf.range(1, (d + 1), delta=1, dtype=x.dtype) * tf.pow(x, 2)
    result = tf.reduce_sum(tf.sin(x) * tf.pow(tf.sin(xi / math.pi), 2 * m), axis=1, keepdims=True)

    return -result


@check_objective_shapes(d=2)
def michalewicz_2(x: TensorType) -> TensorType:
    """
    Convenience function for the 2-dimensional :func:`michalewicz` function with steepness 10.
    :param x: The points at which to evaluate the function, with shape [..., 2].
    :return: The function values at ``x``, with shape [..., 1].
    """
    return michalewicz(x, d=2)


@check_objective_shapes(d=5)
def michalewicz_5(x: TensorType) -> TensorType:
    """
    Convenience function for the 5-dimensional :func:`michalewicz` function with steepness 10.
    :param x: The points at which to evaluate the function, with shape [..., 5].
    :return: The function values at ``x``, with shape [..., 1].
    """
    return michalewicz(x, d=5)


@check_objective_shapes(d=10)
def michalewicz_10(x: TensorType) -> TensorType:
    """
    Convenience function for the 10-dimensional :func:`michalewicz` function with steepness 10.
    :param x: The points at which to evaluate the function, with shape [..., 10].
    :return: The function values at ``x``, with shape [..., 1].
    """
    return michalewicz(x, d=10)


Michalewicz2 = SingleObjectiveTestProblem(
    name="Michalewicz 2",
    objective=michalewicz_2,
    search_space=Box([0.0], [pi]) ** 2,
    minimizers=tf.constant([[2.202906, 1.570796]], tf.float64),
    minimum=tf.constant([-1.8013034], tf.float64),
)
"""Convenience function for the 2-dimensional :func:`michalewicz` function with steepness 10.
Taken from https://arxiv.org/abs/2003.09867"""

Michalewicz5 = SingleObjectiveTestProblem(
    name="Michalewicz 5",
    objective=michalewicz_5,
    search_space=Box([0.0], [pi]) ** 5,
    minimizers=tf.constant([[2.202906, 1.570796, 1.284992, 1.923058, 1.720470]], tf.float64),
    minimum=tf.constant([-4.6876582], tf.float64),
)
"""Convenience function for the 5-dimensional :func:`michalewicz` function with steepness 10.
Taken from https://arxiv.org/abs/2003.09867"""

Michalewicz10 = SingleObjectiveTestProblem(
    name="Michalewicz 10",
    objective=michalewicz_10,
    search_space=Box([0.0], [pi]) ** 10,
    minimizers=tf.constant(
        [
            [
                2.202906,
                1.570796,
                1.284992,
                1.923058,
                1.720470,
                1.570796,
                1.454414,
                1.756087,
                1.655717,
                1.570796,
            ]
        ],
        tf.float64,
    ),
    minimum=tf.constant([-9.6601517], tf.float64),
)
"""Convenience function for the 10-dimensional :func:`michalewicz` function with steepness 10.
Taken from https://arxiv.org/abs/2003.09867"""


def trid(x: TensorType, d: int = 10) -> TensorType:
    """
    The Trid function over :math:`[-d^2, d^2]` for all i=1,...,d. Dimensionality is determined
    by the parameter ``d`` and it has a global minimum. This function has large variation in
    output which makes it challenging for Bayesian optimisation with vanilla Gaussian processes
    with non-stationary kernels. Models that can deal with non-stationarities, such as deep
    Gaussian processes, can be useful for modelling these functions. See :cite:`hebbal2019bayesian`
    and https://www.sfu.ca/~ssurjano/trid.html for details.

    :param x: The points at which to evaluate the function, with shape [..., d].
    :param d: Dimensionality.
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    tf.debugging.assert_greater_equal(d, 2)
    tf.debugging.assert_shapes([(x, (..., d))])

    result = tf.reduce_sum(tf.pow(x - 1, 2), 1, True) - tf.reduce_sum(x[:, 1:] * x[:, :-1], 1, True)
    return result


@check_objective_shapes(d=10)
def trid_10(x: TensorType) -> TensorType:
    """The Trid function with dimension 10.

    :param x: The points at which to evaluate the function, with shape [..., 10].
    :return: The function values at ``x``, with shape [..., 1].
    :raise ValueError (or InvalidArgumentError): If ``x`` has an invalid shape.
    """
    return trid(x, d=10)


Trid10 = SingleObjectiveTestProblem(
    name="Trid 10",
    objective=trid_10,
    search_space=Box([-(10**2)], [10**2]) ** 10,
    minimizers=tf.constant([[i * (10 + 1 - i) for i in range(1, 10 + 1)]], tf.float64),
    minimum=tf.constant([-10 * (10 + 4) * (10 - 1) / 6], tf.float64),
)
"""The Trid function with dimension 10."""



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This module contains common optimizers based on :class:`~tf.optimizers.Optimizer` that can be used
with models. Specific models can also sub-class these optimizers or implement their own, and should
register their loss functions using a :func:`create_loss_function`.
"""

from __future__ import annotations

import copy
from dataclasses import dataclass, field
from functools import singledispatch
from typing import Any, Callable, Iterable, Optional, Tuple, Union

import scipy
import tensorflow as tf
import tensorflow_probability as tfp
from gpflow.keras import tf_keras

from ..data import Dataset
from ..types import TensorType
from ..utils import jit

TrainingData = Union[Tuple[TensorType, TensorType], Iterable[Tuple[TensorType, TensorType]]]
""" Type alias for a batch, or batches, of training data. """

DatasetTransformer = Callable[[Dataset, Optional[int]], TrainingData]
"""
Type alias for a function that converts a :class:`~trieste.data.Dataset` to batches of training
data.
"""

LossClosure = Callable[[], TensorType]
""" Type alias for a loss closure, typically used in optimization. """

OptimizeResult = Union[scipy.optimize.OptimizeResult, None]
"""
Optimization result. TensorFlow optimizer doesn't return any result. For scipy optimizer that is
also commonly used, it is :class:`~scipy.optimize.OptimizeResult`.
"""


@dataclass
class Optimizer:
    """Optimizer for training models with all the training data at once."""

    optimizer: Any
    """
    The underlying optimizer to use. For example, one of the subclasses of
    :class:`~tensorflow.optimizers.Optimizer` could be used. Note that we use a flexible type `Any`
    to allow for various optimizers that specific models might need to use.
    """

    minimize_args: dict[str, Any] = field(default_factory=lambda: {})
    """ The keyword arguments to pass to the :meth:`minimize` method of the :attr:`optimizer`. """

    compile: bool = False
    """ If `True`, the optimization process will be compiled with :func:`~tf.function`. """

    def create_loss(self, model: tf.Module, dataset: Dataset) -> LossClosure:
        """
        Build a loss function for the specified `model` with the `dataset` using a
        :func:`create_loss_function`.

        :param model: The model to build a loss function for.
        :param dataset: The data with which to build the loss function.
        :return: The loss function.
        """
        x = tf.convert_to_tensor(dataset.query_points)
        y = tf.convert_to_tensor(dataset.observations)
        data = (x, y)
        return create_loss_function(model, data, self.compile)

    def optimize(self, model: tf.Module, dataset: Dataset) -> OptimizeResult:
        """
        Optimize the specified `model` with the `dataset`.

        :param model: The model to optimize.
        :param dataset: The data with which to optimize the `model`.
        :return: The return value of the optimizer's :meth:`minimize` method.
        """
        loss_fn = self.create_loss(model, dataset)
        variables = model.trainable_variables
        return self.optimizer.minimize(loss_fn, variables, **self.minimize_args)


@dataclass
class BatchOptimizer(Optimizer):
    """Optimizer for training models with mini-batches of training data."""

    max_iter: int = 100
    """ The number of iterations over which to optimize the model. """

    batch_size: int = 100
    """ The size of the mini-batches. """

    dataset_builder: DatasetTransformer | None = None
    """ A mapping from :class:`~trieste.observer.Observer` data to mini-batches. """

    def create_loss(self, model: tf.Module, dataset: Dataset) -> LossClosure:
        """
        Build a loss function for the specified `model` with the `dataset`.

        :param model: The model to build a loss function for.
        :param dataset: The data with which to build the loss function.
        :return: The loss function.
        """

        def creator_fn(data: TrainingData) -> LossClosure:
            return create_loss_function(model, data, self.compile)

        if self.dataset_builder is None:
            return creator_fn(
                iter(
                    tf.data.Dataset.from_tensor_slices(dataset.astuple())
                    .shuffle(len(dataset))
                    .batch(self.batch_size)
                    .prefetch(tf.data.experimental.AUTOTUNE)
                    .repeat()
                )
            )

        return creator_fn(self.dataset_builder(dataset, self.batch_size))

    def optimize(self, model: tf.Module, dataset: Dataset) -> None:
        """
        Optimize the specified `model` with the `dataset`.

        :param model: The model to optimize.
        :param dataset: The data with which to optimize the `model`.
        """
        loss_fn = self.create_loss(model, dataset)
        variables = model.trainable_variables

        @jit(apply=self.compile)
        def train_fn() -> None:
            self.optimizer.minimize(loss_fn, variables, **self.minimize_args)

        for _ in range(self.max_iter):
            train_fn()

    def __deepcopy__(self, memo: dict[int, object]) -> BatchOptimizer:
        # workaround for https://github.com/tensorflow/tensorflow/issues/58973
        # (keras optimizers not being deepcopyable in TF 2.11 and 2.12)
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if (
                k == "optimizer"
                and isinstance(v, tf_keras.optimizers.Optimizer)
                and hasattr(v, "_distribution_strategy")
            ):
                # avoid copying distribution strategy: reuse it instead
                strategy = v._distribution_strategy
                v._distribution_strategy = None
                try:
                    setattr(result, k, copy.deepcopy(v, memo))
                finally:
                    v._distribution_strategy = strategy
                result.optimizer._distribution_strategy = strategy
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result


@dataclass
class KerasOptimizer:
    """Optimizer wrapper for training models implemented with Keras."""

    optimizer: tf_keras.optimizers.Optimizer
    """ The underlying optimizer to use for training the model. """

    fit_args: dict[str, Any] = field(default_factory=lambda: {})
    """
    The keyword arguments to pass to the ``fit`` method of a :class:`~tf.keras.Model` instance.
    See https://keras.io/api/models/model_training_apis/#fit-method for a list of possible
    arguments in the dictionary.
    """

    loss: Optional[
        Union[
            tf_keras.losses.Loss, Callable[[TensorType, tfp.distributions.Distribution], TensorType]
        ]
    ] = None
    """ Optional loss function for training the model. """

    metrics: Optional[list[tf_keras.metrics.Metric]] = None
    """ Optional metrics for monitoring the performance of the network. """

    def __deepcopy__(self, memo: dict[int, object]) -> KerasOptimizer:
        # workaround for https://github.com/tensorflow/tensorflow/issues/58973
        # (keras optimizers not being deepcopyable in TF 2.11 and 2.12)
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "optimizer" and hasattr(v, "_distribution_strategy"):
                # avoid copying distribution strategy: reuse it instead
                strategy = v._distribution_strategy
                v._distribution_strategy = None
                try:
                    setattr(result, k, copy.deepcopy(v, memo))
                finally:
                    v._distribution_strategy = strategy
                result.optimizer._distribution_strategy = strategy
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result


@singledispatch
def create_loss_function(model: Any, dataset: TrainingData, compile: bool = False) -> LossClosure:
    """
    Generic function for building a loss function for a specified `model` and `dataset`.
    The implementations depends on the type of the model, which should use this function as a
    decorator together with its register method to make a model-specific loss function available.

    :param model: The model to build a loss function for.
    :param dataset: The data with which to build the loss function.
    :param compile: Whether to compile with :func:`tf.function`.
    :return: The loss function.
    """
    raise NotImplementedError(f"Unknown model {model} passed for loss function extraction")



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains builders for GPflow models supported in Trieste. We found the default
configurations used here to work well in most situation, but they should not be taken as
universally good solutions.
"""

from __future__ import annotations

import math
from typing import Callable, Optional, Sequence, Type

import gpflow
import tensorflow as tf
import tensorflow_probability as tfp
from gpflow.kernels import Stationary
from gpflow.models import GPR, SGPR, SVGP, VGP, GPModel

from ...data import Dataset, split_dataset_by_fidelity
from ...space import Box, EncoderFunction, SearchSpace, one_hot_encoded_space, one_hot_encoder
from ...types import TensorType
from ..gpflow.models import GaussianProcessRegression
from ..interfaces import encode_dataset

# NOTE: As a static non-Tensor, this should really be a tf.constant (like the other constants).
# However, changing it breaks serialisation during the expected_improvement.pct.py notebook.
# See https://github.com/secondmind-labs/trieste/issues/833 for details.
KERNEL_LENGTHSCALE = tf.cast(0.2, dtype=gpflow.default_float())
"""
Default value of the kernel lengthscale parameter.
"""


KERNEL_PRIOR_SCALE = tf.constant(1.0, dtype=gpflow.default_float())
"""
Default value of the scaling factor for the kernel lengthscale and variance parameters.
"""


CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE = tf.constant(100.0, dtype=gpflow.default_float())
"""
Default value of the kernel variance parameter for classification models in the noise free case.
"""


CLASSIFICATION_KERNEL_VARIANCE = tf.constant(1.0, dtype=gpflow.default_float())
"""
Default value of the kernel variance parameter for classification models.
"""


MAX_NUM_INDUCING_POINTS = tf.constant(500, dtype=tf.int32)
"""
Default maximum number of inducing points.
"""


NUM_INDUCING_POINTS_PER_DIM = tf.constant(25, dtype=tf.int32)
"""
Default number of inducing points per dimension of the search space.
"""


SIGNAL_NOISE_RATIO_LIKELIHOOD = tf.constant(10, dtype=gpflow.default_float())
"""
Default value used for initializing (noise) variance parameter of the likelihood function.
If user does not specify it, the noise variance is set to maintain the signal to noise ratio
determined by this default value. Signal variance in the kernel is set to the empirical variance.
"""


def build_gpr(
    data: Dataset,
    search_space: Optional[SearchSpace] = None,
    kernel_priors: bool = True,
    likelihood_variance: Optional[float] = None,
    trainable_likelihood: bool = False,
    kernel: Optional[gpflow.kernels.Kernel] = None,
    encoder: EncoderFunction | None = None,
    space_encoder: Callable[[SearchSpace], SearchSpace] | None = None,
) -> GPR:
    """
    Build a :class:`~gpflow.models.GPR` model with sensible initial parameters and
    priors. By default, we use :class:`~gpflow.kernels.Matern52` kernel and
    :class:`~gpflow.mean_functions.Constant` mean function in the model. We found the default
    configuration used here to work well in most situations, but it should not be taken as a
    universally good solution.

    We set priors for kernel hyperparameters by default in order to stabilize model fitting. We
    found the priors below to be highly effective for objective functions defined over the unit
    hypercube. They do seem to work for other search space sizes, but we advise caution when using
    them in such search spaces. Using priors allows for using maximum a posteriori estimate of
    these kernel parameters during model fitting.

    Note that although we scale parameters as a function of the size of the search space, ideally
    inputs should be normalised to the unit hypercube before building a model.

    :param data: Dataset from the initial design, used for estimating the variance of observations.
    :param search_space: Search space for performing Bayesian optimization, used for scaling the
        parameters. Required unless a kernel is passed.
    :param kernel_priors: If set to `True` (default) priors are set for kernel parameters (variance
        and lengthscale).
    :param likelihood_variance: Likelihood (noise) variance parameter can be optionally set to a
        certain value. If left unspecified (default), the noise variance is set to maintain the
        signal to noise ratio of value given by ``SIGNAL_NOISE_RATIO_LIKELIHOOD``, where signal
        variance in the kernel is set to the empirical variance.
    :param trainable_likelihood: If set to `True` Gaussian likelihood parameter is set to
        non-trainable. By default set to `False`.
    :param kernel: The kernel to use in the model, defaults to letting the function set up a
        :class:`~gpflow.kernels.Matern52` kernel.
    :param encoder: Encoder with which to transform the dataset before training. Defaults to
        one_hot_encoder if the search_space is specified.
    :param space_encoder: Encoder with which to transform search_space before generating a kernel.
        Defaults to one_hot_encoded_space.
    :return: A :class:`~gpflow.models.GPR` model.
    """
    if search_space is not None:
        encoder = one_hot_encoder(search_space) if encoder is None else encoder
        space_encoder = one_hot_encoded_space if space_encoder is None else space_encoder
        search_space = space_encoder(search_space)

    if encoder is not None:
        data = encode_dataset(data, encoder)

    empirical_mean, empirical_variance, _ = _get_data_stats(data)

    if kernel is None:
        if search_space is None:
            raise ValueError(
                "'build_gpr' function requires one of 'search_space' or 'kernel' arguments,"
                " but got neither"
            )
        kernel = _get_kernel(empirical_variance, search_space, kernel_priors, kernel_priors)
    mean = _get_mean_function(empirical_mean)

    assert isinstance(kernel, gpflow.kernels.Kernel)
    model = gpflow.models.GPR(data.astuple(), kernel, mean)

    _set_gaussian_likelihood_variance(model, empirical_variance, likelihood_variance)
    gpflow.set_trainable(model.likelihood, trainable_likelihood)

    return model


def build_sgpr(
    data: Dataset,
    search_space: SearchSpace,
    kernel_priors: bool = True,
    likelihood_variance: Optional[float] = None,
    trainable_likelihood: bool = False,
    num_inducing_points: Optional[int] = None,
    trainable_inducing_points: bool = False,
) -> SGPR:
    """
    Build a :class:`~gpflow.models.SGPR` model with sensible initial parameters and
    priors. We use :class:`~gpflow.kernels.Matern52` kernel and
    :class:`~gpflow.mean_functions.Constant` mean function in the model. We found the default
    configuration used here to work well in most situation, but it should not be taken as a
    universally good solution.

    We set priors for kernel hyperparameters by default in order to stabilize model fitting. We
    found the priors below to be highly effective for objective functions defined over the unit
    hypercube. They do seem to work for other search space sizes, but we advise caution when using
    them in such search spaces. Using priors allows for using maximum a posteriori estimate of
    these kernel parameters during model fitting.

    For performance reasons number of inducing points should not be changed during Bayesian
    optimization. Hence, even if the initial dataset is smaller, we advise setting this to a higher
    number. By default inducing points are set to Sobol samples for the continuous search space,
    and simple random samples for discrete or mixed search spaces. This carries
    the risk that optimization gets stuck if they are not trainable, which calls for adaptive
    inducing point selection during the optimization. This functionality will be added to Trieste
    in future.

    Note that although we scale parameters as a function of the size of the search space, ideally
    inputs should be normalised to the unit hypercube before building a model.

    :param data: Dataset from the initial design, used for estimating the variance of observations.
    :param search_space: Search space for performing Bayesian optimization, used for scaling the
        parameters.
    :param kernel_priors: If set to `True` (default) priors are set for kernel parameters (variance
        and lengthscale).
    :param likelihood_variance: Likelihood (noise) variance parameter can be optionally set to a
        certain value. If left unspecified (default), the noise variance is set to maintain the
        signal to noise ratio of value given by ``SIGNAL_NOISE_RATIO_LIKELIHOOD``, where signal
        variance in the kernel is set to the empirical variance.
    :param trainable_likelihood: If set to `True` Gaussian likelihood parameter is set to
        be trainable. By default set to `False`.
    :param num_inducing_points: The number of inducing points can be optionally set to a
        certain value. If left unspecified (default), this number is set to either
        ``NUM_INDUCING_POINTS_PER_DIM``*dimensionality of the search space or value given by
        ``MAX_NUM_INDUCING_POINTS``, whichever is smaller.
    :param trainable_inducing_points: If set to `True` inducing points will be set to
        be trainable. This option should be used with caution. By default set to `False`.
    :return: An :class:`~gpflow.models.SGPR` model.
    """
    empirical_mean, empirical_variance, _ = _get_data_stats(data)

    kernel = _get_kernel(empirical_variance, search_space, kernel_priors, kernel_priors)
    mean = _get_mean_function(empirical_mean)

    inducing_points = gpflow.inducing_variables.InducingPoints(
        _get_inducing_points(search_space, num_inducing_points)
    )

    model = SGPR(data.astuple(), kernel, inducing_points, mean_function=mean)

    _set_gaussian_likelihood_variance(model, empirical_variance, likelihood_variance)
    gpflow.set_trainable(model.likelihood, trainable_likelihood)

    gpflow.set_trainable(model.inducing_variable, trainable_inducing_points)

    return model


def build_vgp_classifier(
    data: Dataset,
    search_space: SearchSpace,
    kernel_priors: bool = True,
    noise_free: bool = False,
    kernel_variance: Optional[float] = None,
) -> VGP:
    """
    Build a :class:`~gpflow.models.VGP` binary classification model with sensible initial
    parameters and priors. We use :class:`~gpflow.kernels.Matern52` kernel and
    :class:`~gpflow.mean_functions.Constant` mean function in the model. We found the default
    configuration used here to work well in most situation, but it should not be taken as a
    universally good solution.

    We set priors for kernel hyperparameters by default in order to stabilize model fitting. We
    found the priors below to be highly effective for objective functions defined over the unit
    hypercube. They do seem to work for other search space sizes, but we advise caution when using
    them in such search spaces. Using priors allows for using maximum a posteriori estimate of
    these kernel parameters during model fitting. In the ``noise_free`` case we do not use prior
    for the kernel variance parameters.

    Note that although we scale parameters as a function of the size of the search space, ideally
    inputs should be normalised to the unit hypercube before building a model.

    :param data: Dataset from the initial design, used for estimating the variance of observations.
    :param search_space: Search space for performing Bayesian optimization, used for scaling the
        parameters.
    :param kernel_priors: If set to `True` (default) priors are set for kernel parameters (variance
        and lengthscale). In the ``noise_free`` case kernel variance prior is not set.
    :param noise_free: If  there is a prior information that the classification problem is a
        deterministic one, this should be set to `True` and kernel variance will be fixed to a
        higher default value ``CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE`` leading to sharper
        classification boundary. In this case prior for the kernel variance parameter is also not
        set. By default set to `False`.
    :param kernel_variance: Kernel variance parameter can be optionally set to a
        certain value. If left unspecified (default), the kernel variance is set to
        ``CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE`` in the ``noise_free`` case and to
        ``CLASSIFICATION_KERNEL_VARIANCE`` otherwise.
    :return: A :class:`~gpflow.models.VGP` model.
    """
    if kernel_variance is not None:
        tf.debugging.assert_positive(kernel_variance)
        variance = tf.cast(kernel_variance, dtype=gpflow.default_float())
    else:
        if noise_free:
            variance = CLASSIFICATION_KERNEL_VARIANCE_NOISE_FREE
        else:
            variance = CLASSIFICATION_KERNEL_VARIANCE

    if noise_free:
        add_prior_to_variance = False
    else:
        add_prior_to_variance = kernel_priors

    model_likelihood = gpflow.likelihoods.Bernoulli()
    kernel = _get_kernel(variance, search_space, kernel_priors, add_prior_to_variance)
    mean = _get_mean_function(tf.constant(0.0, dtype=gpflow.default_float()))

    model = VGP(data.astuple(), kernel, model_likelihood, mean_function=mean)

    gpflow.set_trainable(model.kernel.variance, (not noise_free))

    return model


def build_svgp(
    data: Dataset,
    search_space: SearchSpace,
    classification: bool = False,
    kernel_priors: bool = True,
    likelihood_variance: Optional[float] = None,
    trainable_likelihood: bool = False,
    num_inducing_points: Optional[int] = None,
    trainable_inducing_points: bool = False,
) -> SVGP:
    """
    Build a :class:`~gpflow.models.SVGP` model with sensible initial parameters and
    priors. Both regression and binary classification models are
    available. We use :class:`~gpflow.kernels.Matern52` kernel and
    :class:`~gpflow.mean_functions.Constant` mean function in the model. We found the default
    configuration used here to work well in most situation, but it should not be taken as a
    universally good solution.

    We set priors for kernel hyperparameters by default in order to stabilize model fitting. We
    found the priors below to be highly effective for objective functions defined over the unit
    hypercube. They do seem to work for other search space sizes, but we advise caution when using
    them in such search spaces. Using priors allows for using maximum a posteriori estimate of
    these kernel parameters during model fitting.

    For performance reasons number of inducing points should not be changed during Bayesian
    optimization. Hence, even if the initial dataset is smaller, we advise setting this to a higher
    number. By default inducing points are set to Sobol samples for the continuous search space,
    and simple random samples for discrete or mixed search spaces. This carries
    the risk that optimization gets stuck if they are not trainable, which calls for adaptive
    inducing point selection during the optimization. This functionality will be added to Trieste
    in future.

    Note that although we scale parameters as a function of the size of the search space, ideally
    inputs should be normalised to the unit hypercube before building a model.

    :param data: Dataset from the initial design, used for estimating the variance of observations.
    :param search_space: Search space for performing Bayesian optimization, used for scaling the
        parameters.
    :param classification: If a classification model is needed, this should be set to `True`, in
        which case a Bernoulli likelihood will be used. If a regression model is required, this
        should be set to `False` (default), in which case a Gaussian likelihood is used.
    :param kernel_priors: If set to `True` (default) priors are set for kernel parameters (variance
        and lengthscale).
    :param likelihood_variance: Likelihood (noise) variance parameter can be optionally set to a
        certain value. If left unspecified (default), the noise variance is set to maintain the
        signal to noise ratio of value given by ``SIGNAL_NOISE_RATIO_LIKELIHOOD``, where signal
        variance in the kernel is set to the empirical variance. This argument is ignored in the
        classification case.
    :param trainable_likelihood: If set to `True` likelihood parameter is set to
        be trainable. By default set to `False`. This argument is ignored in the classification
        case.
    :param num_inducing_points: The number of inducing points can be optionally set to a
        certain value. If left unspecified (default), this number is set to either
        ``NUM_INDUCING_POINTS_PER_DIM``*dimensionality of the search space or value given by
        ``MAX_NUM_INDUCING_POINTS``, whichever is smaller.
    :param trainable_inducing_points: If set to `True` inducing points will be set to
        be trainable. This option should be used with caution. By default set to `False`.
    :return: An :class:`~gpflow.models.SVGP` model.
    """
    empirical_mean, empirical_variance, num_data_points = _get_data_stats(data)

    if classification:
        empirical_variance = CLASSIFICATION_KERNEL_VARIANCE
        empirical_mean = tf.constant(0.0, dtype=gpflow.default_float())
        model_likelihood = gpflow.likelihoods.Bernoulli()
    else:
        model_likelihood = gpflow.likelihoods.Gaussian()

    kernel = _get_kernel(empirical_variance, search_space, kernel_priors, kernel_priors)
    mean = _get_mean_function(empirical_mean)

    inducing_points = _get_inducing_points(search_space, num_inducing_points)

    model = SVGP(
        kernel,
        model_likelihood,
        inducing_points,
        mean_function=mean,
        num_data=num_data_points,
    )

    if not classification:
        _set_gaussian_likelihood_variance(model, empirical_variance, likelihood_variance)
        gpflow.set_trainable(model.likelihood, trainable_likelihood)
    gpflow.set_trainable(model.inducing_variable, trainable_inducing_points)

    return model


def _get_data_stats(data: Dataset) -> tuple[TensorType, TensorType, int]:
    empirical_variance = tf.math.reduce_variance(data.observations)
    empirical_mean = tf.math.reduce_mean(data.observations)
    num_data_points = len(data.observations)

    return empirical_mean, empirical_variance, num_data_points


def _get_kernel(
    variance: TensorType,
    search_space: SearchSpace,
    add_prior_to_lengthscale: bool,
    add_prior_to_variance: bool,
) -> gpflow.kernels.Kernel:
    lengthscales = _get_lengthscales(search_space)

    kernel = gpflow.kernels.Matern52(variance=variance, lengthscales=lengthscales)

    if add_prior_to_lengthscale:
        kernel.lengthscales.prior = tfp.distributions.LogNormal(
            tf.math.log(lengthscales), KERNEL_PRIOR_SCALE
        )
    if add_prior_to_variance:
        kernel.variance.prior = tfp.distributions.LogNormal(
            tf.math.log(variance), KERNEL_PRIOR_SCALE
        )

    return kernel


def _get_lengthscales(search_space: SearchSpace) -> TensorType:
    lengthscales = (
        KERNEL_LENGTHSCALE
        * (search_space.upper - search_space.lower)
        * math.sqrt(search_space.dimension)
    )
    search_space_collapsed = tf.equal(search_space.upper, search_space.lower)
    lengthscales = tf.where(
        search_space_collapsed, tf.constant(1.0, dtype=gpflow.default_float()), lengthscales
    )
    return lengthscales


def _get_mean_function(mean: TensorType) -> gpflow.mean_functions.MeanFunction:
    mean_function = gpflow.mean_functions.Constant(mean)

    return mean_function


def _set_gaussian_likelihood_variance(
    model: GPModel, variance: TensorType, likelihood_variance: Optional[float]
) -> None:
    if likelihood_variance is None:
        noise_variance = variance / SIGNAL_NOISE_RATIO_LIKELIHOOD**2
    else:
        tf.debugging.assert_positive(likelihood_variance)
        noise_variance = tf.cast(likelihood_variance, dtype=gpflow.default_float())

    model.likelihood.variance = gpflow.base.Parameter(
        noise_variance, transform=gpflow.utilities.positive(lower=1e-12)
    )


def _get_inducing_points(
    search_space: SearchSpace, num_inducing_points: Optional[int]
) -> TensorType:
    if num_inducing_points is not None:
        tf.debugging.assert_positive(num_inducing_points)
    else:
        num_inducing_points = min(
            MAX_NUM_INDUCING_POINTS, NUM_INDUCING_POINTS_PER_DIM * search_space.dimension
        )
    if isinstance(search_space, Box):
        inducing_points = search_space.sample_sobol(num_inducing_points)
    else:
        inducing_points = search_space.sample(num_inducing_points)

    return inducing_points


def build_multifidelity_autoregressive_models(
    dataset: Dataset,
    num_fidelities: int,
    input_search_space: SearchSpace,
    likelihood_variance: float = 1e-6,
    kernel_priors: bool = False,
    trainable_likelihood: bool = False,
) -> Sequence[GaussianProcessRegression]:
    """
    Build the individual GPR models required for constructing an MultifidelityAutoregressive model
    with `num_fidelities` fidelities.

    :param dataset: Dataset of points with which to initialise the individual models,
        where the final column of the final dimension of the query points contains the fidelity
    :param num_fidelities: Number of fidelities desired for the MultifidelityAutoregressive model
    :param input_search_space: The input search space of the models
    :return: List of initialised GPR models
    """

    # Split data into fidelities
    data = split_dataset_by_fidelity(dataset=dataset, num_fidelities=num_fidelities)

    _validate_multifidelity_data_modellable(data, num_fidelities)

    gprs = [
        GaussianProcessRegression(
            build_gpr(
                data[fidelity],
                input_search_space,
                likelihood_variance=likelihood_variance,
                kernel_priors=kernel_priors,
                trainable_likelihood=trainable_likelihood,
            )
        )
        for fidelity in range(num_fidelities)
    ]

    return gprs


def build_multifidelity_nonlinear_autoregressive_models(
    dataset: Dataset,
    num_fidelities: int,
    input_search_space: SearchSpace,
    kernel_base_class: Type[Stationary] = gpflow.kernels.Matern32,
    kernel_priors: bool = True,
    trainable_likelihood: bool = False,
) -> Sequence[GaussianProcessRegression]:
    """
    Build models for training the trieste.models.gpflow.MultifidelityNonlinearAutoregressive` model

    Builds a basic Matern32 kernel for the lowest fidelity, and the custom kernel described in
    :cite:`perdikaris2017nonlinear` for the higher fidelities, which also have an extra input
    dimension. Note that the initial data that the models with fidelity greater than 0 are
    initialised with contain dummy data in this extra dimension, and so an `update` of the
    `MultifidelityNonlinearAutoregressive` is required to propagate real data through to these
    models.

    :param dataset: The dataset to use to initialise the models
    :param num_fidelities: The number of fidelities to model
    :param input_search_space: the search space, used to initialise the kernel parameters
    :param kernel_base_class: a stationary kernel type
    :param kernel_priors: If set to `True` (default) priors are set for kernel parameters (variance
        and lengthscale).
    :return: gprs: A list containing gprs that can be used for the multifidelity model
    """
    # Split data into fidelities
    data = split_dataset_by_fidelity(dataset=dataset, num_fidelities=num_fidelities)

    _validate_multifidelity_data_modellable(data, num_fidelities)

    # Input dim requires excluding fidelity row
    input_dim = dataset.query_points.shape[1] - 1

    # Create kernels
    kernels = _create_multifidelity_nonlinear_autoregressive_kernels(
        kernel_base_class,
        num_fidelities,
        input_dim,
        input_search_space,
        kernel_priors,
        kernel_priors,
    )

    # Initialise low fidelity GP
    gprs = [
        GaussianProcessRegression(
            build_gpr(
                data[0],
                search_space=input_search_space,  # This isn't actually used when we pass a kernel
                kernel=kernels[0],
                likelihood_variance=1e-6,
                trainable_likelihood=trainable_likelihood,
            )
        )
    ]

    for fidelity in range(1, num_fidelities):
        # Get query points for this fidelity
        qps = data[fidelity].query_points
        samples_column = tf.random.normal([qps.shape[0], 1], dtype=tf.float64)

        augmented_qps = tf.concat([qps, samples_column], axis=1)
        augmented_dataset = Dataset(augmented_qps, data[fidelity].observations)
        gprs.append(
            GaussianProcessRegression(
                build_gpr(
                    augmented_dataset,
                    input_search_space,  # This isn't actually used when we pass a kernel
                    kernel=kernels[fidelity],
                    likelihood_variance=1e-6,
                )
            )
        )

    return gprs


def _validate_multifidelity_data_modellable(data: Sequence[Dataset], num_fidelities: int) -> None:
    if num_fidelities < 2:
        raise ValueError(
            "Invalid number of fidelities to build Multifidelity model for,"
            f" need at least 2 fidelities, got {num_fidelities}"
        )

    for i, fidelity_data in enumerate(data):
        if len(fidelity_data) < 2:
            raise ValueError(
                f"Not enough data to create model for fidelity {i},"
                f" need at least 2 datapoints, got {len(fidelity_data)}"
            )


def _create_multifidelity_nonlinear_autoregressive_kernels(
    kernel_base_class: Type[Stationary],
    n_fidelities: int,
    n_input_dims: int,
    search_space: SearchSpace,
    add_prior_to_lengthscale: bool,
    add_prior_to_variance: bool,
) -> Sequence[Stationary]:
    dims = list(range(n_input_dims + 1))
    lengthscales = _get_lengthscales(search_space)

    scale_lengthscale = 1.0
    kernels = [kernel_base_class(lengthscales=lengthscales)]

    for _ in range(1, n_fidelities):
        interaction_kernel = kernel_base_class(lengthscales=lengthscales, active_dims=dims[:-1])
        scale_kernel = kernel_base_class(lengthscales=scale_lengthscale, active_dims=[dims[-1]])
        bias_kernel = kernel_base_class(lengthscales=lengthscales, active_dims=dims[:-1])
        gpflow.set_trainable(scale_kernel.variance, False)

        if add_prior_to_lengthscale:
            interaction_kernel.lengthscales.prior = tfp.distributions.LogNormal(
                tf.math.log(lengthscales), KERNEL_PRIOR_SCALE
            )
            bias_kernel.lengthscales.prior = tfp.distributions.LogNormal(
                tf.math.log(lengthscales), KERNEL_PRIOR_SCALE
            )
            scale_kernel.lengthscales.prior = tfp.distributions.LogNormal(
                tf.math.log(tf.cast(scale_lengthscale, dtype=gpflow.default_float())),
                KERNEL_PRIOR_SCALE,
            )

        if add_prior_to_variance:
            interaction_kernel.variance.prior = tfp.distributions.LogNormal(
                tf.constant(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
            )
            bias_kernel.variance.prior = tfp.distributions.LogNormal(
                tf.constant(0.0, dtype=gpflow.default_float()), KERNEL_PRIOR_SCALE
            )

        kernels.append(interaction_kernel * scale_kernel + bias_kernel)

    return kernels



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This module contains functionality for optimizing
:data:`~trieste.acquisition.AcquisitionFunction`\ s over :class:`~trieste.space.SearchSpace`\ s.
"""

from __future__ import annotations

from itertools import chain
from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union, cast

import greenlet as gr
import numpy as np
import scipy.optimize as spo
import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import check_shapes

from .. import logging
from ..space import (
    Box,
    CollectionSearchSpace,
    Constraint,
    GeneralDiscreteSearchSpace,
    SearchSpace,
    SearchSpaceType,
    TaggedMultiSearchSpace,
    TaggedProductSearchSpace,
)
from ..types import TensorType
from .interface import AcquisitionFunction

NUM_SAMPLES_MIN: int = 5000
"""
The default minimum number of initial samples for :func:`generate_continuous_optimizer` and
:func:`generate_random_search_optimizer` function, used for determining the number of initial
samples in the multi-start acquisition function optimization.
"""


NUM_SAMPLES_DIM: int = 1000
"""
The default minimum number of initial samples per dimension of the search space for
:func:`generate_continuous_optimizer` function in :func:`automatic_optimizer_selector`, used for
determining the number of initial samples in the multi-start acquisition function optimization.
"""

NUM_RUNS_DIM: int = 10
"""
The default minimum number of optimization runs per dimension of the search space for
:func:`generate_continuous_optimizer` function in :func:`automatic_optimizer_selector`, used for
determining the number of acquisition function optimizations to be performed in parallel.
"""


class FailedOptimizationError(Exception):
    """Raised when an acquisition optimizer fails to optimize"""


AcquisitionOptimizer = Callable[
    [SearchSpaceType, Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]]], TensorType
]
"""
Type alias for a function that returns the single point that maximizes an acquisition function over
a search space or the V points that maximize a vectorized acquisition function (as represented by an
acquisition-int tuple).

If this function receives a search space with points of shape [D] and an acquisition function
with input shape [..., 1, D] output shape [..., 1], the :const:`AcquisitionOptimizer` return shape
should be [1, D].

If instead it receives a search space and a tuple containing the acquisition function and its
vectorization V then the :const:`AcquisitionOptimizer` return shape should be [V, D].
"""


def automatic_optimizer_selector(
    space: SearchSpace, target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]]
) -> TensorType:
    """
    A wrapper around our :const:`AcquisitionOptimizer`s. This class performs
    an :const:`AcquisitionOptimizer` appropriate for the
    problem's :class:`~trieste.space.SearchSpace`.

    :param space: The space of points over which to search, for points with shape [D].
    :param target_func: The function to maximise, with input shape [..., 1, D] and output shape
            [..., 1].
    :return: The batch of points in ``space`` that maximises ``target_func``, with shape [1, D].
    """

    if isinstance(space, GeneralDiscreteSearchSpace):
        return optimize_discrete(space, target_func)

    elif isinstance(space, (Box, CollectionSearchSpace)):
        space_dim = space.dimension
        num_samples = tf.maximum(NUM_SAMPLES_MIN, NUM_SAMPLES_DIM * space_dim)
        num_runs = NUM_RUNS_DIM * space_dim
        return generate_continuous_optimizer(
            num_initial_samples=num_samples,
            num_optimization_runs=num_runs,
        )(space, target_func)

    else:
        raise NotImplementedError(
            f""" No optimizer currently supports acquisition function
                    maximisation over search spaces of type {space}.
                    Try specifying the optimize_random optimizer"""
        )


def _get_max_discrete_points(
    points: TensorType, target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]]
) -> TensorType:
    # check if we need a vectorized optimizer
    if isinstance(target_func, tuple):
        target_func, V = target_func
    else:
        V = 1

    if V < 0:
        raise ValueError(f"vectorization must be positive, got {V}")

    tiled_points = tf.tile(points, [1, V, 1])
    target_func_values = target_func(tiled_points)
    tf.debugging.assert_shapes(
        [(target_func_values, ("_", V))],
        message=(
            f"""
            The result of function target_func has shape
            {tf.shape(target_func_values)}, however, expected a trailing
            dimension of size {V}.
            """
        ),
    )

    best_indices = tf.math.argmax(target_func_values, axis=0)  # [V]
    return tf.gather(tf.transpose(tiled_points, [1, 0, 2]), best_indices, batch_dims=1)  # [V, D]


def optimize_discrete(
    space: GeneralDiscreteSearchSpace,
    target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
) -> TensorType:
    """
    An :const:`AcquisitionOptimizer` for :class:'GeneralDiscreteSearchSpace' spaces.

    When this functions receives an acquisition-integer tuple as its `target_func`,it evaluates
    all the points in the search space for each of the individual V functions making
    up `target_func`.

    :param space: The space of points over which to search, for points with shape [D].
    :param target_func: The function to maximise, with input shape [..., V, D] and output shape
            [..., V].
    :return: The V points in ``space`` that maximises ``target_func``, with shape [V, D].
    """
    points = space.points[:, None, :]
    return _get_max_discrete_points(points, target_func)


InitialPointSampler = Callable[[SearchSpace], Iterable[TensorType]]
"""
Type alias for a function that returns initial point candidates for an optimization.
Candidates are returned in one or more batches, and each batch should have the shape [N, D],
even when N=1.

For simplicity and memory usage, it is recommended to define these as generators. For example,
the following initial point sampler returns both a set of pre-optimized points and 50,000
random samples:

    def sampler(space: SearchSpace) -> Iterable[TensorType]:
        yield pre_optimized_points
        yield space.sample(50_000)

While the following does the same but groups the random samples into batches of size 1,000
to conserve memory:

    def sampler(space: SearchSpace) -> Iterable[TensorType]:
        yield pre_optimized_points
        yield from sample_from_space(50_000, batch_size=1_000)(space)
"""


def sample_from_space(
    num_samples: int,
    batch_size: Optional[int] = None,
    vectorization: int = 1,
) -> InitialPointSampler:
    """
    An initial point sampler that just samples from the search pace.

    :param num_samples: Number of samples to return.
    :param batch_size: If specified, points are return in batches of this size,
        to preserve memory usage.
    :param vectorization: Vectorization of the target function.
    """
    if num_samples <= 0:
        raise ValueError(f"num_samples must be positive, got {num_samples}")

    if isinstance(batch_size, int) and batch_size <= 0:
        raise ValueError(f"batch_size must be positive, got {batch_size}")

    if vectorization <= 0:
        raise ValueError(f"vectorization must be positive, got {vectorization}")

    batch_size_int = batch_size or num_samples

    def sampler(space: SearchSpace) -> Iterable[TensorType]:

        # generate additional points for each vectorization (rather than just replicating them)
        if isinstance(space, TaggedMultiSearchSpace):
            remainder = vectorization % len(space.subspace_tags)
            tf.debugging.assert_equal(
                remainder,
                0,
                message=(
                    f"The vectorization of the target function {vectorization} must be a"
                    f"multiple of the batch shape of initial samples "
                    f"{len(space.subspace_tags)}."
                ),
            )
            multiple = vectorization // len(space.subspace_tags)
        else:
            multiple = vectorization

        for offset in range(0, num_samples, batch_size_int):
            num_batch_samples = min(num_samples - offset, batch_size_int)
            candidates = space.sample(num_batch_samples * multiple)
            candidates = tf.reshape(candidates, [num_batch_samples, vectorization, -1])
            yield candidates

    return sampler


def generate_initial_points(
    num_initial_points: int,
    initial_sampler: InitialPointSampler,
    space: SearchSpace,
    target_func: AcquisitionFunction,
    vectorization: int = 1,
) -> TensorType:
    """
    Return the best starting points for an optimization from those generated by a given sampler.

    :param num_initial_points: Number of best starting points to return.
    :param initial_sampler: Initial point sampler.
    :param space: Search space.
    :param target_func: Target function being optimized.
    :param vectorization: Vectorization of the target function.
    """
    top_fun_values: Optional[TensorType] = None  # [V, num_optimization_runs]
    top_candidates: Optional[TensorType] = None  # [V, num_optimization_runs, D]

    for candidates in initial_sampler(space):
        if tf.rank(candidates) == 3:
            # If samples is a tensor of rank 3, then it is a batch of samples. In this case
            # the vectorization of the target function must be a multiple of the length of the
            # second (batch) dimension.
            remainder = vectorization % tf.shape(candidates)[1]
            tf.debugging.assert_equal(
                remainder,
                tf.constant(0, dtype=remainder.dtype),
                message=(
                    f"""
                    The vectorization of the target function {vectorization} must be a multiple of
                    the batch shape of initial samples {tf.shape(candidates)[1]}.
                    """
                ),
            )
            multiple = vectorization // tf.shape(candidates)[1]
            tiled_candidates = tf.tile(candidates, [1, multiple, 1])  # [samples, V, D]
        else:
            tf.debugging.assert_rank(
                candidates,
                2,
                message=(
                    f"""
                    The initial samples must be a tensor of rank 2, got a tensor of rank
                    {tf.rank(candidates)}.
                    """
                ),
            )
            tiled_candidates = tf.tile(
                candidates[:, None, :], [1, vectorization, 1]
            )  # [samples, V, D]

        target_func_values = target_func(tiled_candidates)  # [samples, V]
        tf.debugging.assert_shapes(
            [(target_func_values, ("_", vectorization))],
            message=(
                f"""
                The result of function target_func has shape
                {tf.shape(target_func_values)}, however, expected a trailing
                dimension of size {vectorization}.
                """
            ),
        )

        # now that we know the output dimension and dtypes, initialize empty top tensors
        if top_candidates is None:
            top_candidates = tf.zeros(
                [vectorization, 0, tf.shape(candidates)[-1]], dtype=candidates.dtype
            )
        if top_fun_values is None:
            top_fun_values = tf.zeros([vectorization, 0], dtype=target_func_values.dtype)

        top_candidates = tf.concat(
            [top_candidates, tf.transpose(tiled_candidates, [1, 0, 2])], 1
        )  # [V, samples+num_initial_points, D]
        top_fun_values = tf.concat(
            [top_fun_values, tf.transpose(target_func_values)], 1
        )  # [V, samples+num_initial_points]

        _, top_k_indices = tf.math.top_k(
            top_fun_values, k=min(num_initial_points, tf.shape(top_fun_values)[-1])
        )  # [V, num_initial_points]

        top_candidates = tf.gather(
            top_candidates, top_k_indices, batch_dims=1
        )  # [V, num_initial_points, D]
        top_fun_values = tf.gather(
            top_fun_values, top_k_indices, batch_dims=1
        )  # [V, num_initial_points]

    if top_candidates is None:
        raise ValueError("No initial point generated!")

    initial_points = tf.transpose(top_candidates, [1, 0, 2])  # [num_initial_points,V,D]
    return initial_points


def generate_continuous_optimizer(
    num_initial_samples: int | InitialPointSampler = NUM_SAMPLES_MIN,
    num_optimization_runs: int = 10,
    num_recovery_runs: int = 10,
    optimizer_args: Optional[dict[str, Any]] = None,
) -> AcquisitionOptimizer[Box | CollectionSearchSpace]:
    """
    Generate a gradient-based optimizer for :class:'Box' and :class:'CollectionSearchSpace'
    spaces and batches of size 1. In the case of a :class:'CollectionSearchSpace', We perform
    gradient-based optimization across all :class:'Box' subspaces, starting from the best location
    found across a sample of `num_initial_samples` random points.

    We advise the user to either use the default `NUM_SAMPLES_MIN` for `num_initial_samples`, or
    `NUM_SAMPLES_DIM` times the dimensionality of the search space, whichever is greater.
    Similarly, for `num_optimization_runs`, we recommend using `NUM_RUNS_DIM` times the
    dimensionality of the search space.

    This optimizer uses Scipy's L-BFGS-B optimizer. We run `num_optimization_runs` separate
    optimizations in parallel, each starting from one of the best `num_optimization_runs` initial
    query points.

    If all `num_optimization_runs` optimizations fail to converge then we run
    `num_recovery_runs` additional runs starting from random locations (also ran in parallel).

    **Note:** using a large number of `num_initial_samples` and `num_optimization_runs` with a
    high-dimensional search space can consume a large amount of CPU memory (RAM).

    :param num_initial_samples: The starting point(s) of the optimization. This can be either
        the number of random samples to use, or a function that given the search space returns
        the points to use. The latter can be used for example to add pre-optimized starting points
        to the random points, as well as to batch point generation to reduce memory usage for
        high-dimensional problems.
    :param num_optimization_runs: The number of separate optimizations to run.
    :param num_recovery_runs: The maximum number of recovery optimization runs in case of failure.
    :param optimizer_args: The keyword arguments to pass to the Scipy L-BFGS-B optimizer.
        Check `minimize` method  of :class:`~scipy.optimize` for details of which arguments
        can be passed. Note that method, jac and bounds cannot/should not be changed.
    :return: The acquisition optimizer.
    """
    if num_optimization_runs <= 0:
        raise ValueError(f"num_optimization_runs must be positive, got {num_optimization_runs}")

    if not callable(num_initial_samples) and num_initial_samples < num_optimization_runs:
        raise ValueError(
            f"""
            num_initial_samples {num_initial_samples} must be at
            least num_optimization_runs {num_optimization_runs}
            """
        )

    if num_recovery_runs < 0:
        raise ValueError(f"num_recovery_runs must be zero or greater, got {num_recovery_runs}")

    def optimize_continuous(
        space: Box | CollectionSearchSpace,
        target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
    ) -> TensorType:
        """
        A gradient-based :const:`AcquisitionOptimizer` for :class:'Box'
        and :class:`CollectionSearchSpace' spaces.

        For :class:'CollectionSearchSpace' we only apply gradient updates to
        its class:'Box' subspaces.

        When this function receives an acquisition-integer tuple as its `target_func`,it
        optimizes each of the individual V functions making up `target_func`, i.e.
        evaluating `num_initial_samples` samples, running `num_optimization_runs` runs, and
        (if necessary) running `num_recovery_runs` recovery run for each of the individual
        V functions.

        :param space: The space over which to search.
        :param target_func: The function to maximise, with input shape [..., V, D] and output shape
                [..., V].
        :return: The V points in ``space`` that maximises``target_func``, with shape [V, D].
        """

        if isinstance(target_func, tuple):  # check if we need a vectorized optimizer
            target_func, V = target_func
        else:
            V = 1

        if V <= 0:
            raise ValueError(f"vectorization must be positive, got {V}")

        initial_sampler = (
            sample_from_space(num_initial_samples, vectorization=V)
            if not callable(num_initial_samples)
            else num_initial_samples
        )

        initial_points = generate_initial_points(
            num_optimization_runs, initial_sampler, space, target_func, V
        )  # [num_optimization_runs,V,D]

        if len(initial_points) < num_optimization_runs:
            raise ValueError(
                f"Not enough initial points generated ({len(initial_points)} "
                f"for {num_optimization_runs} optimization runs)"
            )

        (
            successes,
            fun_values,
            chosen_x,
            nfev,
        ) = _perform_parallel_continuous_optimization(  # [num_optimization_runs, V]
            target_func,
            space,
            initial_points,
            optimizer_args or {},
        )

        successful_optimization = tf.reduce_all(
            tf.reduce_any(successes, axis=0)
        )  # Check that at least one optimization was successful for each function
        total_nfev = tf.reduce_max(nfev)  # acquisition function is evaluated in parallel

        recovery_run = False
        if num_recovery_runs and not successful_optimization:
            # if all optimizations failed for a function then try again from random starts
            random_points = space.sample(num_recovery_runs)
            if tf.rank(random_points) == 3:
                # If samples is a tensor of rank 3, then it is a batch of samples. In this case
                # the vectorization of the target function must be a multiple of the length of the
                # second (batch) dimension.
                remainder = V % tf.shape(random_points)[1]
                tf.debugging.assert_equal(
                    remainder,
                    tf.constant(0, dtype=remainder.dtype),
                    message=(
                        f"""
                        The vectorization of the target function {V} must be a multiple of the batch
                        shape of random samples {tf.shape(random_points)[1]}.
                        """
                    ),
                )
                multiple = V // tf.shape(random_points)[1]
                tiled_random_points = tf.tile(
                    random_points, [1, multiple, 1]  # [num_recovery_runs, V, D]
                )
            else:
                tf.debugging.assert_rank(
                    random_points,
                    2,
                    message=(
                        f"""
                        The random samples must be a tensor of rank 2, got a tensor of rank
                        {tf.rank(random_points)}.
                        """
                    ),
                )
                tiled_random_points = tf.tile(
                    random_points[:, None, :], [1, V, 1]
                )  # [num_recovery_runs, V, D]

            (
                recovery_successes,
                recovery_fun_values,
                recovery_chosen_x,
                recovery_nfev,
            ) = _perform_parallel_continuous_optimization(
                target_func, space, tiled_random_points, optimizer_args or {}
            )

            successes = tf.concat(
                [successes, recovery_successes], axis=0
            )  # [num_optimization_runs + num_recovery_runs, V]
            fun_values = tf.concat(
                [fun_values, recovery_fun_values], axis=0
            )  # [num_optimization_runs + num_recovery_runs, V]
            chosen_x = tf.concat(
                [chosen_x, recovery_chosen_x], axis=0
            )  # [num_optimization_runs + num_recovery_runs, V, D]

            successful_optimization = tf.reduce_all(
                tf.reduce_any(successes, axis=0)
            )  # Check that at least one optimization was successful for each function
            total_nfev += tf.reduce_max(recovery_nfev)
            recovery_run = True

        if not successful_optimization:  # return error if still failed
            raise FailedOptimizationError(
                f"""
                    Acquisition function optimization failed,
                    even after {num_recovery_runs + num_optimization_runs} restarts.
                    """
            )

        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                logging.scalar("spo_af_evaluations", total_nfev)
                if recovery_run:
                    logging.text(
                        "spo_recovery_run",
                        f"Acquisition function optimization failed after {num_optimization_runs} "
                        f"optimization runs, requiring recovery runs",
                    )

                _target_func: AcquisitionFunction = target_func  # make mypy happy

                def improvements() -> tf.Tensor:
                    best_initial_values = tf.math.reduce_max(_target_func(initial_points), axis=0)
                    best_values = tf.math.reduce_max(fun_values, axis=0)
                    improve = best_values - tf.cast(best_initial_values, best_values.dtype)
                    return improve[0] if V == 1 else improve

                if V == 1:
                    logging.scalar("spo_improvement_on_initial_samples", improvements)
                else:
                    logging.histogram("spo_improvements_on_initial_samples", improvements)

        best_run_ids = tf.math.argmax(fun_values, axis=0)  # [V]
        chosen_points = tf.gather(
            tf.transpose(chosen_x, [1, 0, 2]), best_run_ids, batch_dims=1
        )  # [V, D]

        return chosen_points

    return optimize_continuous


def _perform_parallel_continuous_optimization(
    target_func: AcquisitionFunction,
    space: SearchSpace,
    starting_points: TensorType,
    optimizer_args: dict[str, Any],
) -> Tuple[TensorType, TensorType, TensorType, TensorType]:
    """
    A function to perform parallel optimization of our acquisition functions
    using Scipy. We perform L-BFGS-B starting from each of the locations contained
    in `starting_points`, i.e. the number of individual optimization runs is
    given by the leading dimension of `starting_points`.

    To provide a parallel implementation of Scipy's L-BFGS-B that can leverage
    batch calculations with TensorFlow, this function uses the Greenlet package
    to run each individual optimization on micro-threads.

    L-BFGS-B updates for each individual optimization are performed by
    independent greenlets working with Numpy arrays, however, the evaluation
    of our acquisition function (and its gradients) is calculated in parallel
    (for each optimization step) using Tensorflow.

    For :class:'CollectionSearchSpace' we only apply gradient updates to
    its :class:'Box' subspaces, fixing the discrete elements to the best values
    found across the initial random search. To fix these discrete elements, we
    optimize over a continuous :class:'Box' relaxation of the discrete subspaces
    which has equal upper and lower bounds, i.e. we specify an equality constraint
    for this dimension in the scipy optimizer.

    This function also support the maximization of vectorized target functions (with
    vectorization V).

    :param target_func: The function(s) to maximise, with input shape [..., V, D] and
        output shape [..., V].
    :param space: The original search space.
    :param starting_points: The points at which to begin our optimizations of shape
        [num_optimization_runs, V, D]. The leading dimension of
        `starting_points` controls the number of individual optimization runs
        for each of the V target functions.
    :param optimizer_args: Keyword arguments to pass to the Scipy optimizer.
    :return: A tuple containing the failure statuses, maximum values, maximisers and
        number of evaluations for each of our optimizations.
    """

    tf_dtype = starting_points.dtype  # type for communication with Trieste

    num_optimization_runs_per_function = tf.shape(starting_points)[0].numpy()

    V = tf.shape(starting_points)[-2].numpy()  # vectorized batch size
    D = tf.shape(starting_points)[-1].numpy()  # search space dimension
    num_optimization_runs = num_optimization_runs_per_function * V

    vectorized_starting_points = tf.reshape(
        starting_points, [-1, D]
    )  # [num_optimization_runs*V, D]

    def _objective_value(vectorized_x: TensorType) -> TensorType:  # [N, D] -> [N, 1]
        vectorized_x = vectorized_x[:, None, :]  # [N, 1, D]
        x = tf.reshape(vectorized_x, [-1, V, D])  # [N/V, V, D]
        evals = -target_func(x)  # [N/V, V]
        vectorized_evals = tf.reshape(evals, [-1, 1])  # [N, 1]
        return vectorized_evals

    def _objective_value_and_gradient(x: TensorType) -> Tuple[TensorType, TensorType]:
        return tfp.math.value_and_gradient(_objective_value, x)  # [len(x), 1], [len(x), D]

    # Get the bounds for the optimization
    bounds = get_bounds_of_optimization(space, starting_points)

    # Initialize the numpy arrays to be passed to the greenlets
    np_batch_x = np.zeros((num_optimization_runs, tf.shape(starting_points)[-1]), dtype=np.float64)
    np_batch_y = np.zeros((num_optimization_runs,), dtype=np.float64)
    np_batch_dy_dx = np.zeros(
        (num_optimization_runs, tf.shape(starting_points)[-1]), dtype=np.float64
    )

    # Set up child greenlets
    child_greenlets = [ScipyOptimizerGreenlet() for _ in range(num_optimization_runs)]
    vectorized_child_results: List[Union[spo.OptimizeResult, "np.ndarray[Any, Any]"]] = [
        gr.switch(
            vectorized_starting_points[i].numpy(), bounds[i], space.constraints, optimizer_args
        )
        for i, gr in enumerate(child_greenlets)
    ]

    while True:
        all_done = True
        for i, result in enumerate(vectorized_child_results):  # Process results from children.
            if isinstance(result, spo.OptimizeResult):
                continue  # children return a `spo.OptimizeResult` if they are finished
            all_done = False
            assert isinstance(result, np.ndarray)  # or an `np.ndarray` with the query `x` otherwise
            np_batch_x[i, :] = result

        if all_done:
            break

        # Batch evaluate query `x`s from all children.
        batch_x = tf.constant(np_batch_x, dtype=tf_dtype)  # [num_optimization_runs, d]
        batch_y, batch_dy_dx = _objective_value_and_gradient(batch_x)
        np_batch_y = batch_y.numpy().astype("float64")
        np_batch_dy_dx = batch_dy_dx.numpy().astype("float64")

        for i, greenlet in enumerate(child_greenlets):  # Feed `y` and `dy_dx` back to children.
            if greenlet.dead:  # Allow for crashed greenlets
                continue
            vectorized_child_results[i] = greenlet.switch(np_batch_y[i], np_batch_dy_dx[i, :])

    final_vectorized_child_results: List[spo.OptimizeResult] = vectorized_child_results
    vectorized_successes = tf.constant(
        [result.success for result in final_vectorized_child_results]
    )  # [num_optimization_runs]
    vectorized_fun_values = tf.constant(
        [-result.fun for result in final_vectorized_child_results], dtype=tf_dtype
    )  # [num_optimization_runs]
    vectorized_chosen_x = tf.constant(
        [result.x for result in final_vectorized_child_results], dtype=tf_dtype
    )  # [num_optimization_runs, D]
    vectorized_nfev = tf.constant(
        [result.nfev for result in final_vectorized_child_results], dtype=tf_dtype
    )

    # Ensure chosen points satisfy any constraints in the search-space.
    if space.has_constraints:
        is_feasible = space.is_feasible(vectorized_chosen_x)
        vectorized_successes = tf.logical_and(vectorized_successes, is_feasible)

    successes = tf.reshape(vectorized_successes, [-1, V])  # [num_optimization_runs, V]
    fun_values = tf.reshape(vectorized_fun_values, [-1, V])  # [num_optimization_runs, V]
    chosen_x = tf.reshape(vectorized_chosen_x, [-1, V, D])  # [num_optimization_runs, V, D]
    nfev = tf.reshape(vectorized_nfev, [-1, V])  # [num_optimization_runs, V]

    return (successes, fun_values, chosen_x, nfev)


class ScipyOptimizerGreenlet(gr.greenlet):  # type: ignore[misc]
    """
    Worker greenlet that runs a single Scipy L-BFGS-B (by default). Each greenlet performs all the
    optimizer update steps required for an individual optimization. However, the evaluation
    of our acquisition function (and its gradients) is delegated back to the main Tensorflow
    process (the parent greenlet) where evaluations can be made efficiently in parallel.
    """

    def run(
        self,
        start: "np.ndarray[Any, Any]",
        bounds: spo.Bounds,
        constraints: Sequence[Constraint],
        optimizer_args: Optional[dict[str, Any]] = None,
    ) -> spo.OptimizeResult:
        """Greenlet run method."""
        cache_x = start + 1  # Any value different from `start`.
        cache_y: Optional["np.ndarray[Any, Any]"] = None
        cache_dy_dx: Optional["np.ndarray[Any, Any]"] = None

        def value_and_gradient(
            x: "np.ndarray[Any, Any]",
        ) -> Tuple["np.ndarray[Any, Any]", "np.ndarray[Any, Any]"]:
            # Collect function evaluations from parent greenlet
            nonlocal cache_x
            nonlocal cache_y
            nonlocal cache_dy_dx

            if not (cache_x == x).all():
                cache_x[:] = x  # Copy the value of `x`. DO NOT copy the reference.
                # Send `x` to parent greenlet, which will evaluate all `x`s in a batch.
                cache_y, cache_dy_dx = self.parent.switch(cache_x)

            return cast("np.ndarray[Any, Any]", cache_y), cast("np.ndarray[Any, Any]", cache_dy_dx)

        method = "trust-constr" if len(constraints) else "l-bfgs-b"
        optimizer_args = dict(
            {"method": method, "constraints": constraints}, **(optimizer_args or {})
        )
        return spo.minimize(
            lambda x: value_and_gradient(x)[0],
            start,
            jac=lambda x: value_and_gradient(x)[1],
            bounds=bounds,
            **optimizer_args,
        )


def get_bounds_of_box_relaxation_around_point(
    space: TaggedProductSearchSpace, current_point: TensorType
) -> spo.Bounds:
    """
    A function to return the bounds of a continuous relaxation of
    a :class:'TaggedProductSearchSpace' space, i.e. replacing discrete
    spaces with continuous spaces. In particular, all :class:'DiscreteSearchSpace'
    subspaces are replaced with a new :class:'DiscreteSearchSpace' fixed at their
    respective component of the specified 'current_point'. Note that
    all :class:'Box' subspaces remain the same.

    :param space: The original search space.
    :param current_point: The point at which to make the continuous relaxation.
    :return: Bounds for the Scipy optimizer.
    """
    tf.debugging.Assert(isinstance(space, TaggedProductSearchSpace), [tf.constant([])])

    space_with_fixed_discrete = space
    for tag in space.subspace_tags:
        if isinstance(
            space.get_subspace(tag), GeneralDiscreteSearchSpace
        ):  # convert discrete subspaces to box spaces.
            subspace_value = space.get_subspace_component(tag, current_point)
            space_with_fixed_discrete = space_with_fixed_discrete.fix_subspace(tag, subspace_value)
    return spo.Bounds(space_with_fixed_discrete.lower, space_with_fixed_discrete.upper)


@check_shapes(
    "starting_points: [num_optimization_runs, V, D]",
)
def get_bounds_of_optimization(space: SearchSpace, starting_points: TensorType) -> List[spo.Bounds]:
    """
    Returns a list of bounds for all the optimization runs, to be used in the Scipy optimizer.
    The bounds are based on the provided search space and the starting points. The length of the
    list is equal to the total number of individual optimization runs,
    i.e. the number of starting points: `num_optimization_runs` x `V`.

    For each starting point, the bounds are obtained as follows depending on the type of search
    space:
    - For a `TaggedProductSearchSpace`, a "continuous relaxation" of the discrete subspaces is
    built by creating bounds around the point. The discrete components of the created
    search space are fixed at the respective component of the point, and the
    remaining continuous components are set to the original bounds of the search space.
    - For a `TaggedMultiSearchSpace`, the bounds for each point are obtained in a similar
    manner as above, but based potentially on a different subspace for each point;
    instead of sharing a single set of bounds from the common search space.
    The subspaces are assigned to the optimization runs in a round-robin fashion along
    dimension 1 (of size `V`) of the starting points. An error is raised if size `V` is not a
    multiple of the number of subspaces.
    - For other types of search spaces, the original bounds are used for each optimization run.

    :param space: The original search space.
    :param starting_points: The points at which to begin the optimizations, with shape
        `[num_optimization_runs, V, D]`. The leading dimension of `starting_points` controls the
        number of individual optimization runs for each of the `V` starting points or target
        functions.
    :return: A list of bounds for the Scipy optimizer. The length of the list
        is equal to the number of individual optimization runs, i.e. `num_optimization_runs` x `V`.

    For example, for a 2D `TaggedMultiSearchSpace` with two subspaces, each with a continuous
    and a discrete component::

        space = TaggedMultiSearchSpace(
            [
                TaggedProductSearchSpace(
                    [Box([0], [1]), DiscreteSearchSpace([[11], [15], [21], [25], [31], [35]])]
                ),
                TaggedProductSearchSpace(
                    [Box([2], [3]), DiscreteSearchSpace([[13], [17], [23], [27], [33], [37]])]
                )
            ]
        )

    Consider 2 optimization runs per point and a vectorization `V` of 4, for a total of 8
    optimization runs. Given the following starting points::

        starting_points = tf.constant(
            [
                [[10, 11], [12, 13], [14, 15], [16, 17]],
                [[20, 21], [22, 23], [24, 25], [26, 27]],
            ],
            dtype=tf.float64
        )

    The returned list of bounds for the optimization would be::

        [
            spo.Bounds([0, 11], [1, 11]),  # For point at index [0, 0], using subspace 0
            spo.Bounds([2, 13], [3, 13]),  #         //         [0, 1],       //       1
            spo.Bounds([0, 15], [1, 15]),  #         //         [0, 2],       //       0
            spo.Bounds([2, 17], [3, 17]),  #         //         [0, 3],       //       1
            spo.Bounds([0, 21], [1, 21]),  #         //         [1, 0],       //       0
            spo.Bounds([2, 23], [3, 23]),  #         //         [1, 1],       //       1
            spo.Bounds([0, 25], [1, 25]),  #         //         [1, 2],       //       0
            spo.Bounds([2, 27], [3, 27]),  #         //         [1, 3],       //       1
        ]
    """

    num_optimization_runs_per_function, V, D = starting_points.shape
    num_total_optimization_runs = num_optimization_runs_per_function * V

    if isinstance(
        space, TaggedProductSearchSpace
    ):  # build continuous relaxation of discrete subspaces
        vectorized_starting_points = tf.reshape(
            starting_points, [-1, D]
        )  # [num_total_optimization_runs, D]
        bounds = [
            get_bounds_of_box_relaxation_around_point(space, vectorized_starting_points[i : i + 1])
            for i in tf.range(num_total_optimization_runs)
        ]
    elif isinstance(space, TaggedMultiSearchSpace):
        subspaces = [space.get_subspace(tag) for tag in space.subspace_tags]
        # The vectorization `V` of the target function must be a multple of the number of subspaces.
        remainder = V % len(subspaces)
        tf.debugging.assert_equal(
            remainder,
            0,
            message=(
                f"""
                The vectorization of the target function {V} must be a multiple of the number of
                subspaces {len(subspaces)}.
                """
            ),
        )
        multiple = V // len(subspaces)
        subspaces = subspaces * multiple

        # Create a sequence of bounds for each optimization run, per subspace.
        # For product subspaces, we build a continuous relaxation of the discrete subspaces.
        # Otherwise, we use the original bounds.
        bounds = [
            [
                (
                    get_bounds_of_box_relaxation_around_point(ss, starting_points[i : i + 1, j])
                    if isinstance(ss, TaggedProductSearchSpace)
                    else spo.Bounds(ss.lower, ss.upper)
                )
                for j, ss in enumerate(subspaces)
            ]
            for i in tf.range(num_optimization_runs_per_function)
        ]
        bounds = list(chain.from_iterable(bounds))
    else:
        bounds = [spo.Bounds(space.lower, space.upper)] * num_total_optimization_runs

    return bounds


def batchify_joint(
    batch_size_one_optimizer: AcquisitionOptimizer[SearchSpaceType],
    batch_size: int,
) -> AcquisitionOptimizer[SearchSpaceType]:
    """
    A wrapper around our :const:`AcquisitionOptimizer`s. This class wraps a
    :const:`AcquisitionOptimizer` to allow it to jointly optimize the batch elements considered
    by a batch acquisition function.

    :param batch_size_one_optimizer: An optimizer that returns only batch size one, i.e. produces a
            single point with shape [1, D].
    :param batch_size: The number of points in the batch.
    :return: An :const:`AcquisitionOptimizer` that will provide a batch of points with shape [B, D].
    """
    if batch_size <= 0:
        raise ValueError(f"batch_size must be positive, got {batch_size}")

    def optimizer(
        search_space: SearchSpaceType,
        f: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
    ) -> TensorType:
        expanded_search_space = search_space**batch_size  # points have shape [B * D]

        if isinstance(f, tuple):
            raise ValueError(
                "batchify_joint cannot be applied to a vectorized acquisition function"
            )
        af: AcquisitionFunction = f  # type checking can get confused by closure of f

        def target_func_with_vectorized_inputs(
            x: TensorType,
        ) -> TensorType:  # [..., 1, B * D] -> [..., 1]
            return af(tf.reshape(x, x.shape[:-2].as_list() + [batch_size, -1]))

        vectorized_points = batch_size_one_optimizer(  # [1, B * D]
            expanded_search_space, target_func_with_vectorized_inputs
        )
        return tf.reshape(vectorized_points, [batch_size, -1])  # [B, D]

    return optimizer


def batchify_vectorize(
    batch_size_one_optimizer: AcquisitionOptimizer[SearchSpaceType],
    batch_size: int,
) -> AcquisitionOptimizer[SearchSpaceType]:
    """
    A wrapper around our :const:`AcquisitionOptimizer`s. This class wraps a
    :const:`AcquisitionOptimizer` to allow it to optimize batch acquisition functions.

    Unlike :func:`batchify_joint`, :func:`batchify_vectorize` is suitable
    for a :class:`AcquisitionFunction` whose individual batch element can be
    optimized independently (i.e. they can be vectorized).

    :param batch_size_one_optimizer: An optimizer that returns only batch size one, i.e. produces a
            single point with shape [1, D].
    :param batch_size: The number of points in the batch.
    :return: An :const:`AcquisitionOptimizer` that will provide a batch of points with shape [V, D].
    """
    if batch_size <= 0:
        raise ValueError(f"batch_size must be positive, got {batch_size}")

    def optimizer(
        search_space: SearchSpaceType,
        f: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
    ) -> TensorType:
        if isinstance(f, tuple):
            raise ValueError(
                "batchify_vectorize cannot be applied to an already vectorized acquisition function"
            )

        return batch_size_one_optimizer(search_space, (f, batch_size))

    return optimizer


def generate_random_search_optimizer(
    num_samples: int = NUM_SAMPLES_MIN,
) -> AcquisitionOptimizer[SearchSpace]:
    """
    Generate an acquisition optimizer that samples `num_samples` random points across the space.
    The default is to sample at `NUM_SAMPLES_MIN` locations.

    We advise the user to either use the default `NUM_SAMPLES_MIN` for `num_samples`, or
    `NUM_SAMPLES_DIM` times the dimensionality of the search space, whichever is smaller.

    :param num_samples: The number of random points to sample.
    :return: The acquisition optimizer.
    """
    if num_samples <= 0:
        raise ValueError(f"num_samples must be positive, got {num_samples}")

    def optimize_random(
        space: SearchSpace,
        target_func: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
    ) -> TensorType:
        """
        A random search :const:`AcquisitionOptimizer` defined for
        any :class:'SearchSpace' with a :meth:`sample`. If we have a :class:'DiscreteSearchSpace'
        with fewer than `num_samples` points, then we query all the points in the space.

        When this functions receives an acquisition-integer tuple as its `target_func`,it
        optimizes each of the individual V functions making up `target_func`, i.e.
        evaluating `num_samples` samples for each of the individual V functions making up
        target_func.

        :param space: The space over which to search.
        :param target_func: The function to maximise, with input shape [..., V, D] and output shape
                [..., V].
        :return: The V points in ``space`` that maximises ``target_func``, with shape [V, D].
        """
        points = space.sample(num_samples)[:, None, :]
        return _get_max_discrete_points(points, target_func)

    return optimize_random



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains type aliases."""
from typing import Callable, Hashable, Tuple, TypeVar, Union

import tensorflow as tf

TensorType = Union[tf.Tensor, tf.Variable]
"""Type alias for tensor-like types."""

S = TypeVar("S")
"""Unbound type variable."""

T = TypeVar("T")
"""Unbound type variable."""

State = Callable[[S], Tuple[S, T]]
"""
A `State` produces a value of type `T`, given a state of type `S`, and in doing so can update the
state. If the state is updated, it is not updated in-place. Instead, a new state is created. This
is a referentially transparent alternative to mutable state.
"""

Tag = Hashable
"""Type alias for a tag used to label datasets and models."""



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains builders for Keras models supported in Trieste. We found the default
configurations used here to work well in most situation, but they should not be taken as
universally good solutions.
"""

from __future__ import annotations

from typing import Union

from gpflow.keras import tf_keras

from ...data import Dataset
from .architectures import GaussianNetwork, KerasEnsemble
from .utils import get_tensor_spec_from_data


def build_keras_ensemble(
    data: Dataset,
    ensemble_size: int = 5,
    num_hidden_layers: int = 2,
    units: int = 25,
    activation: Union[str, tf_keras.layers.Activation] = "relu",
    independent_normal: bool = False,
) -> KerasEnsemble:
    """
    Builds a simple ensemble of neural networks in Keras where each network has the same
    architecture: number of hidden layers, nodes in hidden layers and activation function.

    Default ensemble size and activation function seem to work well in practice, in regression type
    of problems at least. Number of hidden layers and units per layer should be modified according
    to the dataset size and complexity of the function - the default values seem to work well
    for small datasets common in Bayesian optimization. Using the independent normal is relevant
    only if one is modelling multiple output variables, as it simplifies the distribution by
    ignoring correlations between outputs.

    :param data: Data for training, used for extracting input and output tensor specifications.
    :param ensemble_size: The size of the ensemble, that is, the number of base learners or
        individual neural networks in the ensemble.
    :param num_hidden_layers: The number of hidden layers in each network.
    :param units: The number of nodes in each hidden layer.
    :param activation: The activation function in each hidden layer.
    :param independent_normal: If set to `True` then :class:`~tfp.layers.IndependentNormal` layer
        is used as the output layer. This models outputs as independent, only the diagonal
        elements of the covariance matrix are parametrized. If left as the default `False`,
        then :class:`~tfp.layers.MultivariateNormalTriL` layer is used where correlations
        between outputs are learned as well. Note that this is only relevant for multi-output
        models.
    :return: Keras ensemble model.
    """
    input_tensor_spec, output_tensor_spec = get_tensor_spec_from_data(data)

    hidden_layer_args = []
    for _ in range(num_hidden_layers):
        hidden_layer_args.append({"units": units, "activation": activation})

    networks = [
        GaussianNetwork(
            input_tensor_spec,
            output_tensor_spec,
            hidden_layer_args,
            independent_normal,
        )
        for _ in range(ensemble_size)
    ]
    keras_ensemble = KerasEnsemble(networks)

    return keras_ensemble



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This module registers the GPflow specific loss functions.
"""

from __future__ import annotations

from typing import Any, Callable, Optional

import tensorflow as tf
from gpflow.models import ExternalDataTrainingLossMixin, InternalDataTrainingLossMixin
from tensorflow.python.data.ops.iterator_ops import OwnedIterator as DatasetOwnedIterator

from ..optimizer import LossClosure, TrainingData, create_loss_function


@create_loss_function.register
def _create_loss_function_internal(
    model: InternalDataTrainingLossMixin,
    data: TrainingData,
    compile: bool = False,
) -> LossClosure:
    return model.training_loss_closure(compile=compile)


class _TrainingLossClosureBuilder:
    # A cached, compiled training loss closure builder to avoid having to generate a new
    # closure each time. Stored in a separate class, so we can avoid pickling it.

    def __init__(self) -> None:
        self.closure_builder: Optional[Callable[[TrainingData], LossClosure]] = None

    def __getstate__(self) -> dict[str, Any]:
        return {}

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.closure_builder = None


@create_loss_function.register
def _create_loss_function_external(
    model: ExternalDataTrainingLossMixin,
    data: TrainingData,
    compile: bool = False,
) -> LossClosure:
    if not compile:
        return model.training_loss_closure(data, compile=False)

    # when compiling, we want to avoid generating a new closure every optimization step
    # instead we compile and save a single function that can handle the dynamic data shape
    X, Y = next(data) if isinstance(data, DatasetOwnedIterator) else data

    if not hasattr(model, "_training_loss_closure_builder"):
        setattr(model, "_training_loss_closure_builder", _TrainingLossClosureBuilder())

    builder: _TrainingLossClosureBuilder = getattr(model, "_training_loss_closure_builder")
    if builder.closure_builder is None:
        shape_spec = (
            data.element_spec
            if isinstance(data, DatasetOwnedIterator)
            else (
                tf.TensorSpec([None, *X.shape[1:]], dtype=X.dtype),
                tf.TensorSpec([None, *Y.shape[1:]], dtype=Y.dtype),
            )
        )

        @tf.function(input_signature=shape_spec)
        def training_loss_builder(x: tf.Tensor, y: tf.Tensor) -> tf.Tensor:
            return model.training_loss((x, y))

        def closure_builder(data: TrainingData) -> LossClosure:
            x, y = next(data) if isinstance(data, DatasetOwnedIterator) else data

            def compiled_closure() -> tf.Tensor:
                return training_loss_builder(x, y)

            return compiled_closure

        builder.closure_builder = closure_builder

    return builder.closure_builder((X, Y))



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module is the home of the sampling functionality required by some
of the Trieste's Keras model wrappers.
"""

from __future__ import annotations

from typing import Dict, Optional

import tensorflow as tf

from trieste.utils.misc import ensure_positive

from ...types import TensorType
from ...utils import flatten_leading_dims
from ..interfaces import TrajectoryFunction, TrajectoryFunctionClass, TrajectorySampler
from .interface import DeepEnsembleModel
from .utils import sample_model_index


class DeepEnsembleTrajectorySampler(TrajectorySampler[DeepEnsembleModel]):
    """
    This class builds functions that approximate a trajectory by randomly choosing a network from
    the ensemble and using its predicted means as a trajectory.

    Option `diversify` can be used to increase the diversity in case of optimizing very large
    batches of trajectories. We use quantiles from the approximate Gaussian distribution of
    the ensemble as trajectories, with randomly chosen quantiles approximating a trajectory and
    using a reparametrisation trick to speed up computation. Note that quantiles are not true
    trajectories, so this will likely have some performance costs.
    """

    def __init__(
        self, model: DeepEnsembleModel, diversify: bool = False, seed: Optional[int] = None
    ):
        """
        :param model: The ensemble model to sample from.
        :param diversify: Whether to use quantiles from the approximate Gaussian distribution of
            the ensemble as trajectories (`False` by default). See class docstring for details.
        :param seed: Random number seed to use for trajectory sampling.
        :raise NotImplementedError: If we try to use the model that is not instance of
            :class:`DeepEnsembleModel`.
        """
        if not isinstance(model, DeepEnsembleModel):
            raise NotImplementedError(
                f"EnsembleTrajectorySampler only works with DeepEnsembleModel models, that support "
                f"ensemble_size and ensemble_distributions methods; "
                f"received {model!r}"
            )

        super().__init__(model)

        self._model = model
        self._diversify = diversify
        self._seed = seed or int(tf.random.uniform(shape=(), maxval=10000, dtype=tf.int32))

    def __repr__(self) -> str:
        """"""
        return f"{self.__class__.__name__}({self._model!r}"

    def get_trajectory(self) -> TrajectoryFunction:
        """
        Generate an approximate function draw (trajectory) from the ensemble.

        :return: A trajectory function representing an approximate trajectory
            from the model, taking an input of shape `[N, B, D]` and returning shape `[N, B, L]`.
        """
        return deep_ensemble_trajectory(self._model, self._diversify, self._seed)

    def update_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Update a :const:`TrajectoryFunction` to reflect an update in its
        underlying :class:`DeepEnsembleModel` and resample accordingly.

        Here we rely on the underlying models being updated and we only resample the trajectory.

        :param trajectory: The trajectory function to be resampled.
        :return: The new trajectory function updated for a new model
        """
        tf.debugging.Assert(isinstance(trajectory, deep_ensemble_trajectory), [tf.constant([])])
        trajectory.resample()  # type: ignore
        return trajectory

    def resample_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Efficiently resample a :const:`TrajectoryFunction` in-place to avoid function retracing
        with every new sample.

        :param trajectory: The trajectory function to be resampled.
        :return: The new resampled trajectory function.
        """
        tf.debugging.Assert(isinstance(trajectory, deep_ensemble_trajectory), [tf.constant([])])
        trajectory.resample()  # type: ignore
        return trajectory


class deep_ensemble_trajectory(TrajectoryFunctionClass):
    """
    Generate an approximate function draw (trajectory) by randomly choosing a batch B of
    networks from the ensemble and using their predicted means as trajectories.

    Option `diversify` can be used to increase the diversity in case of optimizing very large
    batches of trajectories. We use quantiles from the approximate Gaussian distribution of
    the ensemble as trajectories, with randomly chosen quantiles approximating a trajectory and
    using a reparametrisation trick to speed up computation. Only epistemic uncertainty is taken
    into account in sampling. Note that quantiles are not true trajectories, so this will likely
    have some performance costs.
    """

    def __init__(self, model: DeepEnsembleModel, diversify: bool, seed: Optional[int] = None):
        """
        :param model: The model of the objective function.
        :param diversify: Whether to use samples from final probabilistic layer as trajectories
            or mean predictions.
        :param seed: Optional RNG seed.
        """
        self._model = model
        self._diversify = diversify
        self._ensemble_size = self._model.ensemble_size
        self._seed = seed

        self._initialized = tf.Variable(False, trainable=False)
        self._batch_size = tf.Variable(0, dtype=tf.int32, trainable=False)

        if self._diversify:
            self._eps = tf.Variable(
                tf.zeros([0, 0], dtype=model.dtype), shape=[None, None], trainable=False
            )
        else:
            self._indices = tf.Variable(
                tf.zeros([0], dtype=tf.int32), shape=[None], trainable=False
            )

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:  # [N, B, D] -> [N, B, L]
        """
        Call trajectory function. Note that we are flattening the batch dimension and
        doing a forward pass with each network in the ensemble with the whole batch. This is
        somewhat wasteful, but is necessary given the underlying ``KerasEnsemble`` network
        model.
        """
        if not self._initialized:  # work out desired batch size from input
            self._batch_size.assign(tf.shape(x)[-2])  # B
            self.resample()  # sample network indices/quantiles
            self._initialized.assign(True)

        tf.debugging.assert_equal(
            tf.shape(x)[-2],
            self._batch_size,
            message=f"""
            This trajectory only supports batch sizes of {self._batch_size}.
            If you wish to change the batch size you must get a new trajectory
            by calling the get_trajectory method of the trajectory sampler.
            """,
        )
        flat_x, unflatten = flatten_leading_dims(x)  # [N*B, D]

        if self._diversify:
            predicted_means, predicted_vars = self._model.predict(flat_x)  # ([N*B, L], [N*B, L])
            predicted_vars = ensure_positive(predicted_vars)
            predictions = predicted_means + tf.sqrt(predicted_vars) * tf.tile(
                self._eps, [tf.shape(x)[0], 1]
            )  # [N*B, L]
            return unflatten(predictions)  # [N, B, L]
        else:
            ensemble_distributions = self._model.ensemble_distributions(flat_x)
            predicted_means = tf.convert_to_tensor([dist.mean() for dist in ensemble_distributions])
            predictions = tf.gather(predicted_means, self._indices)  # [B, N*B, L]

            tensor_predictions = tf.map_fn(unflatten, predictions)  # [B, N, B, L]

            # here we select simultaneously networks and batch dimension according to batch indices
            # this is needed because we compute a whole batch with each network
            batch_index = tf.range(self._batch_size)
            indices = tf.stack([batch_index, batch_index], axis=1)
            batch_predictions = tf.gather_nd(
                tf.transpose(tensor_predictions, perm=[0, 2, 1, 3]), indices
            )  # [B,N]

            return tf.transpose(batch_predictions, perm=[1, 0, 2])  # [N, B, L]

    def resample(self) -> None:
        """
        Efficiently resample network indices in-place, without retracing.
        """
        if self._seed:
            self._seed += 1  # increment operation seed

        if self._diversify:
            self._eps.assign(
                tf.random.normal(
                    shape=(self._batch_size, self._model.num_outputs),
                    dtype=self._model.dtype,
                    seed=self._seed,
                )
            )  # [B]
        else:
            self._indices.assign(
                sample_model_index(self._ensemble_size, self._batch_size, seed=self._seed)
            )  # [B]

    def get_state(self) -> Dict[str, TensorType]:
        """
        Return internal state variables.
        """
        state = {
            "initialized": self._initialized,
            "batch_size": self._batch_size,
        }
        if self._diversify:
            state["eps"] = self._eps
        else:
            state["indices"] = self._indices

        return state



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This module contains functions and classes for Pareto based multi-objective optimization. """
from __future__ import annotations

try:
    import cvxpy as cp
except ImportError:  # pragma: no cover (tested but not by coverage)
    cp = None
import numpy as np
import tensorflow as tf

from ...types import TensorType
from .dominance import non_dominated
from .partition import prepare_default_non_dominated_partition_bounds


class Pareto:
    """
    A :class:`Pareto` constructs a Pareto set.
    Stores a Pareto set and calculates hypervolume of the Pareto set given a
    specified reference point
    """

    def __init__(self, observations: TensorType, already_non_dominated: bool = False):
        """
        :param observations: The observations for all objectives, with shape [N, D].
        :param already_non_dominated: Whether the observations are already non dominated
        :raise ValueError (or InvalidArgumentError): If ``observations`` has an invalid shape.
        """
        tf.debugging.assert_rank(observations, 2)
        tf.debugging.assert_greater_equal(tf.shape(observations)[-1], 2)

        if not already_non_dominated:
            self.front = non_dominated(observations)[0]
        else:
            self.front = observations

    def hypervolume_indicator(self, reference: TensorType) -> TensorType:
        """
        Calculate the hypervolume indicator based on self.front and a reference point
        The hypervolume indicator is the volume of the dominated region.

        :param reference: a reference point to use, with shape [D].
            Defines the upper bound of the hypervolume.
            Should be equal or bigger than the anti-ideal point of the Pareto set.
            For comparing results across runs, the same reference point must be used.
        :return: hypervolume indicator, if reference point is less than all of the front
            in any dimension, the hypervolume indicator will be zero.
        :raise ValueError (or `tf.errors.InvalidArgumentError`): If ``reference`` has an invalid
            shape.
        :raise ValueError (or `tf.errors.InvalidArgumentError`): If ``self.front`` is empty
            (which can happen if the concentration point is too strict so no frontier
            exists after the screening)
        """
        if tf.equal(tf.size(self.front), 0):
            raise ValueError("empty front cannot be used to calculate hypervolume indicator")

        helper_anti_reference = tf.reduce_min(self.front, axis=0) - tf.ones(
            shape=1, dtype=self.front.dtype
        )
        lower, upper = prepare_default_non_dominated_partition_bounds(
            reference, self.front, helper_anti_reference
        )
        non_dominated_hypervolume = tf.reduce_sum(tf.reduce_prod(upper - lower, 1))
        hypervolume_indicator = (
            tf.reduce_prod(reference - helper_anti_reference) - non_dominated_hypervolume
        )
        return hypervolume_indicator

    def sample_diverse_subset(
        self,
        sample_size: int,
        allow_repeats: bool = True,
        bounds_delta_scale_factor: float = 0.2,
        bounds_min_delta: float = 1e-9,
    ) -> tuple[TensorType, TensorType]:
        """
        Sample a set of diverse points from the Pareto set using
        Hypervolume Sharpe-Ratio Indicator

        :param sample_size: The number of points to sample from the Pareto front
        :param allow_repeats: Whether the sample may contain repeats
        :param bounds_delta_scale_factor: The factor by which to grow the distance
            between extrema when calculating lower and upper bounds
        :param bounds_min_delta: The minimum value of the distance between extrema
        :return: sample: Tensor of query points selected in the sample
                 and sample_ids: Tensor of indices of points selected from the Pareto set
        """

        if cp is None:
            raise ImportError(
                "Pareto.sample method requires cvxpy, "
                "this can be installed via `pip install trieste[qhsri]`"
            )

        if bounds_delta_scale_factor < 0:
            raise ValueError(
                "bounds_delta_scale_factor should be non-negative,"
                f" got {bounds_delta_scale_factor}"
            )
        if bounds_delta_scale_factor < 0:
            raise ValueError("bounds_delta_min should be non-negative, got {bounds_min_delta}")

        front_size, _ = self.front.shape

        if (front_size < sample_size) and allow_repeats is False:
            raise ValueError(
                f"Tried to sample {sample_size} points from a Pareto"
                f" set of size {front_size}, please ensure sample size is smaller than"
                " Pareto set size."
            )

        lower_bound, upper_bound = self._get_bounds(bounds_delta_scale_factor, bounds_min_delta)

        p = self._calculate_p_matrix(lower_bound, upper_bound)

        # Calculate q matrix
        p_diag = np.expand_dims(np.diagonal(p), axis=1)
        q = p - np.dot(p_diag, np.transpose(p_diag))

        x_star = self._find_x_star(q, p)

        if allow_repeats:
            samples, sample_ids = self._choose_batch_with_repeats(x_star, sample_size)
        else:
            samples, sample_ids = self._choose_batch_no_repeats(x_star, sample_size)

        return samples, sample_ids

    def _choose_batch_with_repeats(
        self, x_star: TensorType, sample_size: int
    ) -> tuple[TensorType, TensorType]:
        # Calculate number of times each point is sampled
        n_times_sampled = x_star * sample_size

        # Round each number to an int
        n_times_sampled = np.round(n_times_sampled)

        # Check difference in number of samples and required sample size
        n_samples_difference = sample_size - int(np.sum(n_times_sampled))

        if n_samples_difference < 0:
            # We need to randomly remove samples
            # Get indices of point that were to be sampled >0 times
            non_zero_indices = np.flatnonzero(n_times_sampled)
            # Choose indices to decrement
            samples_to_decr = np.random.choice(non_zero_indices, size=-n_samples_difference)
            # Decrement indices
            for idx in samples_to_decr:
                n_times_sampled[idx] -= 1
        elif n_samples_difference > 0:
            # We need to randomly add samples
            samples_to_incr = np.random.choice(
                np.arange(len(n_times_sampled)), size=n_samples_difference
            )
            for idx in samples_to_incr:
                n_times_sampled[idx] += 1

        # Create a list of the sample indices
        sample_ids = []
        for idx, repeats in enumerate(n_times_sampled):
            for _ in range(int(repeats)):
                sample_ids.append(idx)
        # Convert to array for indexing and return
        sample_ids_array = np.array(sample_ids)

        # Create batch with each of the selected samples
        samples = np.array(self.front)[sample_ids_array]

        return samples, sample_ids_array

    def _choose_batch_no_repeats(
        self, x_star: TensorType, sample_size: int
    ) -> tuple[TensorType, TensorType]:
        front_size = self.front.shape[0]

        # Create id array to keep track of points
        id_arr = np.expand_dims(np.arange(front_size), axis=1)

        # Stitch id array, x_star and the front together
        stitched_array = np.concatenate([id_arr, x_star, np.array(self.front)], axis=1)

        # Sort array by x_star descending
        sorted_array = stitched_array[stitched_array[:, 1].argsort()[::-1]]

        samples = sorted_array[:sample_size, 2:]
        sample_ids = sorted_array[:sample_size, 0].astype(int)

        return samples, sample_ids

    def _find_x_star(self, q: TensorType, p: TensorType) -> TensorType:
        front_size = self.front.shape[0]

        p_diag = np.expand_dims(np.diagonal(p), axis=1)

        # Solve quadratic program for y*
        P = cp.atoms.affine.wraps.psd_wrap(q)
        G = np.eye(front_size)
        h = np.zeros(front_size)
        A = np.transpose(p_diag)
        b = np.ones(1)

        # Define and solve the CVXPY problem.
        y = cp.Variable(front_size)
        prob = cp.Problem(cp.Minimize((1 / 2) * cp.quad_form(y, P)), [G @ y >= h, A @ y == b])
        prob.solve()

        y_star = y.value

        # Calculate x*
        x_star = np.expand_dims(y_star, axis=1) / np.sum(y_star)

        return x_star

    def _calculate_p_matrix(self, lower_bound: TensorType, upper_bound: TensorType) -> TensorType:
        front_size, front_dims = self.front.shape

        p = np.zeros([front_size, front_size])

        # Calculate denominator value for p matrix elements
        denominator: float = 1
        for i in range(front_dims):
            if upper_bound[i] - lower_bound[i] == 0:
                raise ValueError(
                    "Pareto set has identical upper and lower bounds"
                    " in a dimension, you can avoid this by setting a "
                    "nonzero value for bounds_min_delta"
                )
            denominator *= upper_bound[i] - lower_bound[i]

        # Fill entries of p
        for i in range(front_size):
            for j in range(front_size):
                pij = 1
                for k in range(front_dims):
                    pij *= upper_bound[k] - max(self.front[i, k], self.front[j, k])
                p[i, j] = pij

        p = p / denominator

        return p

    def _get_bounds(
        self, delta_scaling_factor: float, min_delta: float
    ) -> tuple[TensorType, TensorType]:
        # Find min and max for each dimension in the front
        dim_mins = np.min(self.front, axis=0)
        dim_maxes = np.max(self.front, axis=0)
        # Calculate the deltas to add to the min/max to get the upper and lower bounds
        deltas = ((dim_maxes - dim_mins) * delta_scaling_factor) + min_delta
        # Calculate the bounds
        lower_bound = dim_mins - deltas
        upper_bound = dim_maxes + deltas

        return lower_bound, upper_bound


def get_reference_point(
    observations: TensorType,
) -> TensorType:
    """
    Default reference point calculation method that calculates the reference
    point according to a Pareto front extracted from set of observations.

    :param observations: observations referred to calculate the reference
        point, with shape [..., N, D]
    :return: a reference point to use, with shape [..., D].
    :raise ValueError: If ``observations`` is empty
    """
    if tf.equal(tf.size(observations), 0):
        raise ValueError("empty observations cannot be used to calculate reference point")

    front = Pareto(observations).front
    f = tf.math.reduce_max(front, axis=-2) - tf.math.reduce_min(front, axis=-2)
    return tf.math.reduce_max(front, axis=-2) + 2 * f / tf.cast(tf.shape(front)[-2], f.dtype)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains utility functions for acquisition functions.
"""
from typing import Callable, Tuple

import tensorflow as tf
from tensorflow_probability import distributions as tfd

from ...types import TensorType

# =============================================================================
# Multivariate Normal CDF
# =============================================================================


class MultivariateNormalCDF:
    def __init__(
        self,
        sample_size: int,
        dim: int,
        dtype: tf.DType,
        num_sobol_skip: int = 0,
    ) -> None:
        """Builds the cumulative density function of the multivariate Gaussian
        using the Genz approximation detailed in :cite:`genz2016numerical`.

        This is a Monte Carlo approximation which is more accurate than a naive
        Monte Carlo estimate of the expected improvent. In order to use
        reparametrised samples, the helper accepts a tensor of samples, and the
        callable uses these fixed samples whenever it is called.

        :param samples_size: int, number of samples to use.
        :param dim: int, dimension of the multivariate Gaussian.
        :param dtype: tf.DType, data type to use for calculations.
        :param num_sobol_skip: int, number of sobol samples to skip.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_positive(dim)

        self._S = sample_size
        self._Q = dim
        self._dtype = dtype

        self._num_sobol_skip = num_sobol_skip

    def _standard_normal_cdf_and_inverse_cdf(
        self,
        dtype: tf.DType,
    ) -> Tuple[Callable[[TensorType], TensorType], Callable[[TensorType], TensorType]]:
        """Returns two callables *Phi* and *iPhi*, which compute the cumulative
        density function and inverse cumulative density function of a standard
        univariate Gaussian.

        :param dtype: The data type to use, either tf.float32 or tf.float64.
        :returns Phi, iPhi: Cumulative and inverse cumulative density functions.
        """

        normal = tfd.Normal(
            loc=tf.zeros(shape=(), dtype=dtype),
            scale=tf.ones(shape=(), dtype=dtype),
        )
        Phi: Callable[[TensorType], TensorType] = normal.cdf
        iPhi: Callable[[TensorType], TensorType] = normal.quantile

        return Phi, iPhi

    def _get_update_indices(self, B: int, S: int, Q: int, q: int) -> TensorType:
        """Returns indices for updating a tensor using tf.tensor_scatter_nd_add,
        for use within the _mvn_cdf function, for computing the cumulative density
        function of a multivariate Gaussian. The indices *idx* returned are such
        that the following operation

            idx = get_update_indices(B, S, Q, q)
            tensor = tf.tensor_scatter_nd_add(tensor, idx, update)

        is equivalent to the numpy operation

            tensor = tensor[:, :, q] + update

        where *tensor* is a tensor of shape (B, S, Q).

        :param B: First dim. of tensor for which the indices are generated.
        :param S: Second dim. of tensor for which the indices are generated.
        :param Q: Third dim. of tensor for which the indices are generated.
        :param q: Index of tensor along fourth dim. to which the update is applied.
        """

        idxB = tf.tile(tf.range(B, dtype=tf.int32)[:, None, None], (1, S, 1))
        idxS = tf.tile(tf.range(S, dtype=tf.int32)[None, :, None], (B, 1, 1))
        idxQ = tf.tile(tf.convert_to_tensor(q)[None, None, None], (B, S, 1))

        idx = tf.concat([idxB, idxS, idxQ], axis=-1)

        return idx

    def __call__(
        self,
        x: TensorType,
        mean: TensorType,
        cov: TensorType,
        jitter: float = 1e-6,
    ) -> TensorType:
        """Computes the cumulative density function of the multivariate
        Gaussian using the Genz approximation.

        :param x: Tensor of shape (B, Q), batch of points to evaluate CDF at.
        :param mean: Tensor of shape (B, Q), batch of means.
        :param covariance: Tensor of shape (B, Q, Q), batch of covariances.
        :param jitter: float, jitter to use in the Cholesky factorisation.
        :returns mvn_cdf: Tensor of shape (B,), CDF values.
        """

        # Unpack batch size
        B = x.shape[0]
        tf.debugging.assert_positive(B)

        # Check shapes of input tensors
        tf.debugging.assert_shapes(
            [
                (x, (B, self._Q)),
                (mean, (B, self._Q)),
                (cov, (B, self._Q, self._Q)),
            ]
        )

        # Identify data type to use for all calculations
        dtype = mean.dtype

        # Compute Cholesky factors
        jitter = jitter * tf.eye(self._Q, dtype=dtype)[None, :, :]
        C = tf.linalg.cholesky(cov + jitter)  # (B, Q, Q)

        # Rename samples and limits for brevity
        w = tf.math.sobol_sample(
            dim=self._Q,
            num_results=self._S,
            dtype=self._dtype,
            skip=self._num_sobol_skip,
        )  # (S, Q)
        b = x - mean  # (B, Q)

        # Initialise transformation variables
        e = tf.zeros(shape=(B, self._S, self._Q), dtype=dtype)
        f = tf.zeros(shape=(B, self._S, self._Q), dtype=dtype)
        y = tf.zeros(shape=(B, self._S, self._Q), dtype=dtype)

        # Initialise standard normal for computing CDFs
        Phi, iPhi = self._standard_normal_cdf_and_inverse_cdf(dtype=dtype)

        # Get update indices for convenience later
        idx = self._get_update_indices(B=B, S=self._S, Q=self._Q, q=0)

        # Slice out common tensors
        b0 = b[:, None, 0]
        C0 = C[:, None, 0, 0] + 1e-12

        # Compute transformation variables at the first step
        e_update = tf.tile(Phi(b0 / C0), (1, self._S))  # (B, S)
        e = tf.tensor_scatter_nd_add(e, idx, e_update)
        f = tf.tensor_scatter_nd_add(f, idx, e_update)

        for i in tf.range(1, self._Q):
            # Update y tensor
            y_update = iPhi(1e-6 + (1 - 2e-6) * w[None, :, i - 1] * e[:, :, i - 1])
            y = tf.tensor_scatter_nd_add(y, idx, y_update)

            # Slice out common tensors
            bi = b[:, None, i]
            Ci_ = C[:, None, i, :i]
            Cii = C[:, None, i, i] + 1e-12
            yi = y[:, :, :i]

            # Compute indices to update d, e and f tensors
            idx = self._get_update_indices(B=B, S=self._S, Q=self._Q, q=i)

            # Update e tensor
            e_update = Phi((bi - tf.reduce_sum(Ci_ * yi, axis=-1)) / Cii)
            e = tf.tensor_scatter_nd_add(e, idx, e_update)

            # Update f tensor
            f_update = e[:, :, i] * f[:, :, i - 1]
            f = tf.tensor_scatter_nd_add(f, idx, f_update)

        mvn_cdf = tf.reduce_mean(f[:, :, -1], axis=-1)

        return mvn_cdf



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This package contains the primary interface for deep Gaussian process models. It also contains a
number of :class:`TrainableProbabilisticModel` wrappers for GPflux-based models.
"""

from .builders import build_vanilla_deep_gp
from .interface import GPfluxPredictor
from .models import DeepGaussianProcess
from .sampler import (
    DeepGaussianProcessDecoupledLayer,
    DeepGaussianProcessDecoupledTrajectorySampler,
    DeepGaussianProcessReparamSampler,
    ResampleableDecoupledDeepGaussianProcessFeatureFunctions,
    dgp_feature_decomposition_trajectory,
)
